import warnings

# A collection of deprecated roboFab methods.
# Those methods are added to keep scripts and code compatible.


class RemovedError(Exception):
    """Exception for things removed from FontParts that were in RoboFab"""


# ========
# = base =
# ========


class RemovedBase:
    def setParent(self, parent):
        objName = self.__class__.__name__.replace("Removed", "")
        raise RemovedError(f"'{objName}.setParent()'")


class DeprecatedBase:
    def update(self):
        objName = self.__class__.__name__.replace("Deprecated", "")
        warnings.warn(
            f"'{objName}.update': use {objName}.changed()", DeprecationWarning
        )
        self.changed()

    def setChanged(self):
        objName = self.__class__.__name__.replace("Deprecated", "")
        warnings.warn(
            f"'{objName}.setChanged': use {objName}.changed()", DeprecationWarning
        )
        self.changed()


# ==================
# = transformation =
# ==================


class DeprecatedTransformation:
    def move(self, *args, **kwargs):
        objName = self.__class__.__name__.replace("Deprecated", "")
        warnings.warn(f"'{objName}.move()': use {objName}.moveBy()", DeprecationWarning)
        self.moveBy(*args, **kwargs)

    def translate(self, *args, **kwargs):
        objName = self.__class__.__name__.replace("Deprecated", "")
        warnings.warn(
            f"'{objName}.translate()': use {objName}.moveBy()", DeprecationWarning
        )
        self.moveBy(*args, **kwargs)

    def scale(self, *args, **kwargs):
        objName = self.__class__.__name__.replace("Deprecated", "")
        warnings.warn(
            f"'{objName}.scale()': use {objName}.scaleBy()", DeprecationWarning
        )
        if "center" in kwargs:
            kwargs["origin"] = kwargs["center"]
            del kwargs["center"]
        self.scaleBy(*args, **kwargs)

    def rotate(self, *args, **kwargs):
        objName = self.__class__.__name__.replace("Deprecated", "")
        warnings.warn(
            f"'{objName}.rotate()': use {objName}.rotateBy()", DeprecationWarning
        )
        if "offset" in kwargs:
            kwargs["origin"] = kwargs["offset"]
            del kwargs["offset"]
        self.rotateBy(*args, **kwargs)

    def transform(self, *args, **kwargs):
        objName = self.__class__.__name__.replace("Deprecated", "")
        warnings.warn(
            f"'{objName}.transform()': use {objName}.transformBy()", DeprecationWarning
        )
        self.transformBy(*args, **kwargs)

    def skew(self, *args, **kwargs):
        objName = self.__class__.__name__.replace("Deprecated", "")
        warnings.warn(f"'{objName}.skew()': use {objName}.skewBy()", DeprecationWarning)
        if "offset" in kwargs:
            kwargs["origin"] = kwargs["offset"]
            del kwargs["offset"]
        self.skewBy(*args, **kwargs)


# =========
# = Point =
# =========


class RemovedPoint(RemovedBase):
    @staticmethod
    def select(state=True):
        raise RemovedError("'Point.select'")


class DeprecatedPoint(DeprecatedBase, DeprecatedTransformation):
    def _generateIdentifier(self):
        warnings.warn(
            "'Point._generateIdentifier()': use 'Point._getIdentifier()'",
            DeprecationWarning,
        )
        return self._getIdentifier()

    def generateIdentifier(self):
        warnings.warn(
            "'Point.generateIdentifier()': use 'Point.getIdentifier()'",
            DeprecationWarning,
        )
        return self.getIdentifier()

    def getParent(self):
        warnings.warn("'Point.getParent()': use 'Point.contour'", DeprecationWarning)
        return self.contour


# ==========
# = BPoint =
# ==========


class RemovedBPoint(RemovedBase):
    @staticmethod
    def select(state=True):
        raise RemovedError("'BPoint.select'")


class DeprecatedBPoint(DeprecatedBase, DeprecatedTransformation):
    def _generateIdentifier(self):
        warnings.warn(
            "'BPoint._generateIdentifier()': use 'BPoint._getIdentifier()'",
            DeprecationWarning,
        )
        return self._getIdentifier()

    def generateIdentifier(self):
        warnings.warn(
            "'BPoint.generateIdentifier()': use 'BPoint.getIdentifier()'",
            DeprecationWarning,
        )
        return self.getIdentifier()

    def getParent(self):
        warnings.warn("'BPoint.getParent()': use 'BPoint.contour'", DeprecationWarning)
        return self.contour


# ==========
# = Anchor =
# ==========


class RemovedAnchor(RemovedBase):
    @staticmethod
    def draw(pen):
        raise RemovedError("'Anchor.draw': UFO3 is not drawing anchors into pens")

    @staticmethod
    def drawPoints(pen):
        raise RemovedError(
            "'Anchor.drawPoints': UFO3 is not drawing anchors into point pens"
        )


class DeprecatedAnchor(DeprecatedBase, DeprecatedTransformation):
    def _generateIdentifier(self):
        warnings.warn(
            "'Anchor._generateIdentifier()': use 'Anchor._getIdentifier()'",
            DeprecationWarning,
        )
        return self._getIdentifier()

    def generateIdentifier(self):
        warnings.warn(
            "'Anchor.generateIdentifier()': use 'Anchor.getIdentifier()'",
            DeprecationWarning,
        )
        return self.getIdentifier()

    def getParent(self):
        warnings.warn("'Anchor.getParent()': use 'Anchor.glyph'", DeprecationWarning)
        return self.glyph


# =============
# = Component =
# =============


class RemovedComponent(RemovedBase):
    pass


class DeprecatedComponent(DeprecatedBase):
    def _get_box(self):
        warnings.warn("'Component.box': use Component.bounds", DeprecationWarning)
        return self.bounds

    box = property(_get_box, doc="Deprecated Component.box")

    def _generateIdentifier(self):
        warnings.warn(
            ("'Component._generateIdentifier()': use 'Component._getIdentifier()'"),
            DeprecationWarning,
        )
        return self._getIdentifier()

    def generateIdentifier(self):
        warnings.warn(
            ("'Component.generateIdentifier()': use 'Component.getIdentifier()'"),
            DeprecationWarning,
        )
        return self.getIdentifier()

    def getParent(self):
        warnings.warn(
            "'Component.getParent()': use 'Component.glyph'", DeprecationWarning
        )
        return self.glyph

    def move(self, *args, **kwargs):
        warnings.warn("'Component.move()': use Component.moveBy()", DeprecationWarning)
        self.moveBy(*args, **kwargs)

    def translate(self, *args, **kwargs):
        warnings.warn(
            "'Component.translate()': use Component.moveBy()", DeprecationWarning
        )
        self.moveBy(*args, **kwargs)

    def rotate(self, *args, **kwargs):
        warnings.warn(
            "'Component.rotate()': use Component.rotateBy()", DeprecationWarning
        )
        if "offset" in kwargs:
            kwargs["origin"] = kwargs["offset"]
            del kwargs["offset"]
        self.rotateBy(*args, **kwargs)

    def transform(self, *args, **kwargs):
        warnings.warn(
            "'Component.transform()': use Component.transformBy()", DeprecationWarning
        )
        self.transformBy(*args, **kwargs)

    def skew(self, *args, **kwargs):
        warnings.warn("'Component.skew()': use Component.skewBy()", DeprecationWarning)
        if "offset" in kwargs:
            kwargs["origin"] = kwargs["offset"]
            del kwargs["offset"]
        self.skewBy(*args, **kwargs)


# ===========
# = Segment =
# ===========


class RemovedSegment(RemovedBase):
    @staticmethod
    def insertPoint(point):
        raise RemovedError("Segment.insertPoint()")

    @staticmethod
    def removePoint(point):
        raise RemovedError("Segment.removePoint()")


class DeprecatedSegment(DeprecatedBase, DeprecatedTransformation):
    def getParent(self):
        warnings.warn(
            "'Segment.getParent()': use 'Segment.contour'", DeprecationWarning
        )
        return self.contour


# ===========
# = Contour =
# ===========


class RemovedContour(RemovedBase):
    pass


class DeprecatedContour(DeprecatedBase, DeprecatedTransformation):
    def _get_box(self):
        warnings.warn("'Contour.box': use Contour.bounds", DeprecationWarning)
        return self.bounds

    box = property(_get_box, doc="Deprecated Contour.box")

    def reverseContour(self):
        warnings.warn(
            "'Contour.reverseContour()': use 'Contour.reverse()'", DeprecationWarning
        )
        self.reverse()

    def _generateIdentifier(self):
        warnings.warn(
            "'Contour._generateIdentifier()': use 'Contour._getIdentifier()'",
            DeprecationWarning,
        )
        return self._getIdentifier()

    def generateIdentifier(self):
        warnings.warn(
            "'Contour.generateIdentifier()': use 'Contour.getIdentifier()'",
            DeprecationWarning,
        )
        return self.getIdentifier()

    def _generateIdentifierforPoint(self, point):
        warnings.warn(
            (
                "'Contour._generateIdentifierforPoint()': use "
                "'Contour._getIdentifierforPoint()'"
            ),
            DeprecationWarning,
        )
        return self._getIdentifierForPoint(point)

    def generateIdentifierforPoint(self, point):
        warnings.warn(
            (
                "'Contour.generateIdentifierforPoint()': use "
                "'Contour.getIdentifierForPoint()'"
            ),
            DeprecationWarning,
        )
        return self.getIdentifierForPoint(point)

    def getParent(self):
        warnings.warn("'Contour.getParent()': use 'Contour.glyph'", DeprecationWarning)
        return self.glyph


# =========
# = Glyph =
# =========


class RemovedGlyph(RemovedBase):
    @staticmethod
    def center(padding=None):
        raise RemovedError("'Glyph.center()'")

    @staticmethod
    def clearVGuides():
        raise RemovedError("'Glyph.clearVGuides()': use Glyph.clearGuidelines()")

    @staticmethod
    def clearHGuides():
        raise RemovedError("'Glyph.clearHGuides()': use Glyph.clearGuidelines()")


class DeprecatedGlyph(DeprecatedBase, DeprecatedTransformation):
    def _get_mark(self):
        warnings.warn("'Glyph.mark': use Glyph.markColor", DeprecationWarning)
        return self.markColor

    def _set_mark(self, value):
        warnings.warn("'Glyph.mark': use Glyph.markColor", DeprecationWarning)
        self.markColor = value

    mark = property(_get_mark, _set_mark, doc="Deprecated Mark color")

    def _get_box(self):
        warnings.warn("'Glyph.box': use Glyph.bounds", DeprecationWarning)
        return self.bounds

    box = property(_get_box, doc="Deprecated Glyph.box")

    def getAnchors(self):
        warnings.warn("'Glyph.getAnchors()': use Glyph.anchors", DeprecationWarning)
        return self.anchors

    def getComponents(self):
        warnings.warn(
            "'Glyph.getComponents()': use Glyph.components", DeprecationWarning
        )
        return self.components

    def getParent(self):
        warnings.warn("'Glyph.getParent()': use 'Glyph.font'", DeprecationWarning)
        return self.font

    def readGlyphFromString(self, glifData):
        warnings.warn(
            ("'Glyph.readGlyphFromString()': use 'Glyph.loadFromGLIF()'"),
            DeprecationWarning,
        )
        return self.loadFromGLIF(glifData)

    def writeGlyphToString(self, glyphFormatVersion=2):
        warnings.warn(
            ("'Glyph.writeGlyphToString()': use 'Glyph.dumpToGLIF()'"),
            DeprecationWarning,
        )
        return self.dumpToGLIF(glyphFormatVersion)


# =============
# = Guideline =
# =============


class RemovedGuideline(RemovedBase):
    pass


class DeprecatedGuideline(DeprecatedBase, DeprecatedTransformation):
    def _generateIdentifier(self):
        warnings.warn(
            ("'Guideline._generateIdentifier()': use 'Guideline._getIdentifier()'"),
            DeprecationWarning,
        )
        return self._getIdentifier()

    def generateIdentifier(self):
        warnings.warn(
            ("'Guideline.generateIdentifier()': use 'Guideline.getIdentifier()'"),
            DeprecationWarning,
        )
        return self.getIdentifier()

    def getParent(self):
        warnings.warn(
            ("'Guideline.getParent()': use 'Guideline.glyph' or 'Guideline.font'"),
            DeprecationWarning,
        )
        glyph = self.glyph
        if glyph is not None:
            return glyph
        return self.font


# =======
# = Lib =
# =======


class RemovedLib(RemovedBase):
    pass


class DeprecatedLib:
    def getParent(self):
        warnings.warn(
            "'Lib.getParent()': use 'Lib.glyph' or 'Lib.font'", DeprecationWarning
        )
        glyph = self.glyph
        if glyph is not None:
            return glyph
        return self.font

    def setChanged(self):
        warnings.warn("'Lib.setChanged': use Lib.changed()", DeprecationWarning)
        self.changed()


# ==========
# = Groups =
# ==========


class RemovedGroups(RemovedBase):
    pass


class DeprecatedGroups:
    def getParent(self):
        warnings.warn("'Groups.getParent()': use 'Groups.font'", DeprecationWarning)
        return self.font

    def setChanged(self):
        warnings.warn("'Groups.setChanged': use Groups.changed()", DeprecationWarning)
        self.changed()


# ===========
# = Kerning =
# ===========


class RemovedKerning:
    @staticmethod
    def setParent(parent):
        raise RemovedError("'Kerning.setParent()'")

    @staticmethod
    def swapNames(swaptable):
        raise RemovedError("Kerning.swapNames()")

    @staticmethod
    def getLeft(glyphName):
        raise RemovedError("Kerning.getLeft()")

    @staticmethod
    def getRight(glyphName):
        raise RemovedError("Kerning.getRight()")

    @staticmethod
    def getExtremes():
        raise RemovedError("Kerning.getExtremes()")

    @staticmethod
    def add(value):
        raise RemovedError("Kerning.add()")

    @staticmethod
    def minimize(minimum=10):
        raise RemovedError("Kerning.minimize()")

    @staticmethod
    def importAFM(path, clearExisting=True):
        raise RemovedError("Kerning.importAFM()")

    @staticmethod
    def getAverage():
        raise RemovedError("Kerning.getAverage()")

    @staticmethod
    def combine(kerningDicts, overwriteExisting=True):
        raise RemovedError("Kerning.combine()")

    @staticmethod
    def eliminate(
        leftGlyphsToEliminate=None, rightGlyphsToEliminate=None, analyzeOnly=False
    ):
        raise RemovedError("Kerning.eliminate()")

    @staticmethod
    def occurrenceCount(glyphsToCount):
        raise RemovedError("Kerning.occurrenceCount()")

    @staticmethod
    def implodeClasses(leftClassDict=None, rightClassDict=None, analyzeOnly=False):
        raise RemovedError("Kerning.implodeClasses()")

    @staticmethod
    def explodeClasses(leftClassDict=None, rightClassDict=None, analyzeOnly=False):
        raise RemovedError("Kerning.explodeClasses()")


class DeprecatedKerning:
    def setChanged(self):
        warnings.warn("'Kerning.setChanged': use Kerning.changed()", DeprecationWarning)
        self.changed()

    def getParent(self):
        warnings.warn("'Kerning.getParent()': use 'Kerning.font'", DeprecationWarning)
        return self.font


# ========
# = Info =
# ========


class RemovedInfo(RemovedBase):
    pass


class DeprecatedInfo(DeprecatedBase):
    def getParent(self):
        warnings.warn("'Info.getParent()': use 'Info.font'", DeprecationWarning)
        return self.font


# =========
# = Image =
# =========


class RemovedImage(RemovedBase):
    pass


class DeprecatedImage(DeprecatedBase):
    def getParent(self):
        warnings.warn("'Image.getParent()': use 'Image.glyph'", DeprecationWarning)
        return self.glyph


# ============
# = Features =
# ============


class RemovedFeatures(RemovedBase):
    @staticmethod
    def round():
        raise RemovedError("'Features.round()'")


class DeprecatedFeatures(DeprecatedBase):
    def getParent(self):
        warnings.warn("'Features.getParent()': use 'Features.font'", DeprecationWarning)
        return self.font


# =========
# = Layer =
# =========


class RemovedLayer(RemovedBase):
    pass


class DeprecatedLayer(DeprecatedBase):
    def getParent(self):
        warnings.warn("'Layer.getParent()': use 'Layer.font'", DeprecationWarning)
        return self.font


# ========
# = Font =
# ========


class RemovedFont(RemovedBase):
    @staticmethod
    def getParent():
        raise RemovedError("'Font.getParent()'")

    @staticmethod
    def generateGlyph(*args, **kwargs):
        raise RemovedError("'Font.generateGlyph()'")

    @staticmethod
    def compileGlyph(*args, **kwargs):
        raise RemovedError("'Font.compileGlyph()'")

    @staticmethod
    def getGlyphNameToFileNameFunc():
        raise RemovedError("'Font.getGlyphNameToFileNameFunc()'")


class DeprecatedFont(DeprecatedBase):
    def _get_fileName(self):
        warnings.warn(
            "'Font.fileName': use os.path.basename(Font.path)", DeprecationWarning
        )
        return self.path

    fileName = property(_get_fileName, doc="Deprecated Font.fileName")

    def getWidth(self, glyphName):
        warnings.warn(
            "'Font.getWidth(): use Font[glyphName].width'", DeprecationWarning
        )
        return self[glyphName].width

    def getGlyph(self, glyphName):
        warnings.warn("'Font.getGlyph(): use Font[glyphName]'", DeprecationWarning)
        return self[glyphName]

    def _get_selection(self):
        warnings.warn(
            "'Font.selection: use Font.selectedGlyphNames'", DeprecationWarning
        )
        return self.selectedGlyphNames

    def _set_selection(self, glyphNames):
        warnings.warn(
            "'Font.selection: use Font.selectedGlyphNames'", DeprecationWarning
        )
        self.selectedGlyphNames = glyphNames

    selection = property(
        _get_selection, _set_selection, doc="Deprecated Font.selection"
    )



import unittest
import collections
import tempfile
import os
import shutil
from unittest.mock import PropertyMock, patch
from fontTools.ufoLib import DEFAULT_LAYER_NAME


class TestFont(unittest.TestCase):
    # ----
    # repr
    # ----
    def test_reprContents(self):
        font, _ = self.objectGenerator("font")
        font.info.familyName = "testFamily"
        font.info.styleName = "testStyle"
        result = font._reprContents()
        self.assertEqual(len(result), 1)
        self.assertIn("'testFamily testStyle'", result)

    def test_reprContents_path(self):
        font, _ = self.objectGenerator("font")
        font.info.familyName = "testFamily"
        font.info.styleName = "testStyle"
        with patch.object(type(font), "path", new_callable=PropertyMock) as mock_path:
            mock_path.return_value = "path/to/font.ufo"
            result = font._reprContents()
            self.assertEqual(len(result), 2)
            self.assertIn("'testFamily testStyle'", result)
            self.assertIn("path='path/to/font.ufo'", result)

    # ------
    # Layers
    # ------

    def getFont_layers(self):
        font, _ = self.objectGenerator("font")
        for name in "ABCD":
            font.newLayer("layer " + name)
        return font

    def test_getLayer_unknown(self):
        font = self.getFont_layers()
        with self.assertRaises(ValueError):
            font.getLayer("There is no layer with this name.")

    def test_newLayer_layer_exists(self):
        font = self.getFont_layers()
        font.newLayer("testLayer")
        self.assertIn("testLayer", font.layerOrder)
        font.newLayer("testLayer")
        self.assertIn("testLayer", font.layerOrder)

    def test_newLayer_colored_layer_exists(self):
        font = self.getFont_layers()
        font.newLayer("testLayer", (1, 1, 1, 1))
        self.assertIn("testLayer", font.layerOrder)
        layer = font.newLayer("testLayer", (0, 1, 1, 0))
        self.assertIn("testLayer", font.layerOrder)
        self.assertEqual(layer.color, (0, 1, 1, 0))

    def test_removeLayer(self):
        font, _ = self.objectGenerator("font")
        layer = font.newLayer("testLayer")
        self.assertIn(layer, font.layers)
        font.removeLayer("testLayer")
        self.assertNotIn(layer, font.layers)
        with self.assertRaises(ValueError):
            font.removeLayer("testLayer")

    def test_insertLayer(self):
        font, _ = self.objectGenerator("font")
        layer, _ = self.objectGenerator("layer")
        layer.name = "oldName"
        insertedLayer = font.insertLayer(layer, name="newName")
        self.assertEqual(insertedLayer.name, "newName")
        self.assertIn("newName", font.layerOrder)

    def test_insertLayer_name_none(self):
        font, _ = self.objectGenerator("font")
        layer, _ = self.objectGenerator("layer")
        layer.name = "testLayer"
        font.insertLayer(layer, name=None)
        self.assertIn(layer.name, font.layerOrder)

    def test_insertLayer_name_exists(self):
        font, _ = self.objectGenerator("font")
        existingLayer = font.newLayer("existingLayer")
        self.assertIn(existingLayer.name, font.layerOrder)
        newLayer, _ = self.objectGenerator("layer")
        newLayer.name = "newLayer"
        font.insertLayer(newLayer, name=existingLayer.name)
        self.assertIn(existingLayer.name, font.layerOrder)

    def test_dublicateLayer(self):
        font, _ = self.objectGenerator("font")
        existingLayer = font.newLayer("testLayer")
        self.assertIn(existingLayer.name, font.layerOrder)
        duplicatedLayer = font.duplicateLayer(existingLayer.name, "duplicateLayer")
        self.assertIn(duplicatedLayer.name, font.layerOrder)

    def test_duplicateLayer_layerName_missing(self):
        font, _ = self.objectGenerator("font")
        with self.assertRaises(ValueError):
            font.duplicateLayer("missingName", "duplicateLayer")

    def test_duplicateLayer_newLayerName_exists(self):
        font, _ = self.objectGenerator("font")
        existingLayer = font.newLayer("testLayer")
        duplicateLayer = font.newLayer("duplicateLayer")
        with self.assertRaises(ValueError):
            font.duplicateLayer(existingLayer.name, duplicateLayer.name)

    def test_swapLayerNames(self):
        font, _ = self.objectGenerator("font")
        layer1 = font.newLayer("layer1")
        layer2 = font.newLayer("layer2")
        font.swapLayerNames(layer1.name, layer2.name)
        self.assertEqual(layer1.name, "layer2")
        self.assertEqual(layer2.name, "layer1")

    def test_swapLayerNames_names_missing(self):
        font, _ = self.objectGenerator("font")
        layer = font.newLayer("testLayer")
        testCases = [("missingLayer", layer.name), (layer.name, "missingLayer")]
        for name1, name2 in testCases:
            with self.assertRaises(ValueError):
                font.swapLayerNames(name1, name2)

    # ------
    # Glyphs
    # ------

    def getFont_glyphs(self):
        font, _ = self.objectGenerator("font")
        for name in "ABCD":
            font.newGlyph(name)
        return font

    # glyphOrder

    def test_glyphOrder(self):
        font = self.getFont_glyphs()
        expectedGlyphOrder = ["A", "B", "C", "D"]
        self.assertEqual(font.glyphOrder, tuple(expectedGlyphOrder))
        # reverse the excepected glyph order and set it
        expectedGlyphOrder.reverse()
        font.glyphOrder = expectedGlyphOrder
        self.assertEqual(font.glyphOrder, tuple(expectedGlyphOrder))
        # add a glyph
        expectedGlyphOrder.append("newGlyph")
        font.newGlyph("newGlyph")
        self.assertEqual(font.glyphOrder, tuple(expectedGlyphOrder))
        # remove a glyph
        expectedGlyphOrder.remove("newGlyph")
        del font["newGlyph"]
        self.assertEqual(font.glyphOrder, tuple(expectedGlyphOrder))
        # insert a glyph, where the glyph is at the beginning of the glyph order
        glyph, _ = self.objectGenerator("glyph")
        font["D"] = glyph
        self.assertEqual(font.glyphOrder, tuple(expectedGlyphOrder))
        # insert a glyph, where the glyph is at the end of the glyph order
        glyph, _ = self.objectGenerator("glyph")
        font["A"] = glyph
        self.assertEqual(font.glyphOrder, tuple(expectedGlyphOrder))

    # len

    def test_len_initial(self):
        font = self.getFont_glyphs()
        self.assertEqual(len(font), 4)

    def test_len_two_layers(self):
        font = self.getFont_glyphs()
        layer = font.newLayer("test")
        layer.newGlyph("X")
        self.assertEqual(len(font), 4)

    # insert glyphs

    def test_insertGlyph(self):
        font, _ = self.objectGenerator("font")
        glyph, _ = self.objectGenerator("glyph")
        font.insertGlyph(glyph, "inserted1")
        self.assertIn("inserted1", font)
        font["inserted2"] = glyph
        self.assertIn("inserted2", font)
        font.newGlyph("A")
        glyph.unicode = 123
        font["A"] = glyph
        self.assertEqual(font["A"].unicode, 123)

    # add glyphs

    def test_newGlyph_rename(self):
        layer = self.getFont_glyphs()

        layer.newGlyph("A", clear=False, rename=True)
        layer.newGlyph("A", clear=False, rename=True)

        self.assertIn("A", layer)
        self.assertIn("A.1", layer)
        self.assertIn("A.2", layer)

    # ----------
    # Guidelines
    # ----------

    def getFont_guidelines(self):
        font, _ = self.objectGenerator("font")
        font.appendGuideline((1, 2), 0, "Test Guideline 1")
        font.appendGuideline((3, 4), 90, "Test Guideline 2")
        return font

    def test_appendGuideline_valid_object(self):
        font, _ = self.objectGenerator("font")
        src, _ = self.objectGenerator("guideline")
        src.position = (1, 2)
        src.angle = 123
        src.name = "test"
        src.color = (1, 1, 1, 1)
        src.getIdentifier()
        dst = font.appendGuideline(guideline=src)
        self.assertNotEqual(src, dst)
        self.assertEqual(src.position, dst.position)
        self.assertEqual(src.angle, dst.angle)
        self.assertEqual(src.name, dst.name)
        self.assertEqual(src.color, dst.color)
        self.assertEqual(src.identifier, dst.identifier)

    def test_appendGuideline_position(self):
        font = self.getFont_guidelines()
        initialLength = len(font.guidelines)
        font.appendGuideline(position=(0, 0))
        self.assertEqual(len(font.guidelines), initialLength + 1)
        self.assertEqual(font.guidelines[-1].position, (0, 0))

    def test_appendGuideline_overrides(self):
        font = self.getFont_guidelines()
        guideline, _ = self.objectGenerator("guideline")
        guideline.position = (0, 0)
        initialLength = len(font.guidelines)
        font.appendGuideline(
            position=(10, 10),
            angle=1,
            name="test",
            color=(1, 1, 1, 1),
            guideline=guideline,
        )
        self.assertEqual(len(font.guidelines), initialLength + 1)
        self.assertEqual(font.guidelines[-1].position, (10, 10))
        self.assertEqual(font.guidelines[-1].angle, 1)
        self.assertEqual(font.guidelines[-1].name, "test")
        self.assertEqual(font.guidelines[-1].color, (1, 1, 1, 1))

    def test_appendGuideline_raises(self):
        font = self.getFont_guidelines()
        with self.assertRaises(ValueError):
            font.appendGuideline(position=None)

    def test_removeGuideline_index(self):
        font = self.getFont_guidelines()
        initialLength = len(font.guidelines)
        font.removeGuideline(0)
        self.assertEqual(len(font.guidelines), initialLength - 1)

    def test_removeGuideline_index_out_of_range(self):
        font = self.getFont_guidelines()
        with self.assertRaises(ValueError):
            font.appendGuideline(position=None)

    def test_clearGuideline(self):
        font = self.getFont_guidelines()
        font.clearGuidelines()
        self.assertEqual(font.guidelines, ())

    # ----
    # flatKerning
    # ----

    def test_flatKerning(self):
        font = self.getFont_glyphs()
        # glyph, glyph kerning
        font.kerning["A", "V"] = -100
        font.kerning["V", "A"] = -200
        expected = {("A", "V"): -100, ("V", "A"): -200}
        self.assertEqual(font.getFlatKerning(), expected)
        # add some groups
        font.groups["public.kern1.O"] = ["O", "Ograve"]
        font.groups["public.kern2.O"] = ["O", "Ograve"]
        # group, group kerning
        font.kerning["public.kern1.O", "public.kern2.O"] = -50
        expected = {
            ("O", "O"): -50,
            ("Ograve", "O"): -50,
            ("O", "Ograve"): -50,
            ("Ograve", "Ograve"): -50,
            ("A", "V"): -100,
            ("V", "A"): -200,
        }
        self.assertEqual(font.getFlatKerning(), expected)
        # glyph, group exception
        font.kerning["O", "public.kern2.O"] = -30
        expected = {
            ("O", "O"): -30,
            ("Ograve", "O"): -50,
            ("O", "Ograve"): -30,
            ("Ograve", "Ograve"): -50,
            ("A", "V"): -100,
            ("V", "A"): -200,
        }
        self.assertEqual(font.getFlatKerning(), expected)
        # glyph, glyph exception
        font.kerning["O", "Ograve"] = -70
        expected = {
            ("O", "O"): -30,
            ("Ograve", "O"): -50,
            ("O", "Ograve"): -70,
            ("Ograve", "Ograve"): -50,
            ("A", "V"): -100,
            ("V", "A"): -200,
        }
        self.assertEqual(font.getFlatKerning(), expected)

    # -------
    # TempLib
    # -------

    def test_font_tempLib(self):
        font, _ = self.objectGenerator("font")
        tempLib = font.tempLib
        tempLib["test.key"] = "test.value"
        self.assertEqual(tempLib["test.key"], "test.value")
        self.assertEqual(tempLib.font, font)

    # ----
    # Hash
    # ----

    def test_hash_same_object(self):
        font_one = self.getFont_glyphs()
        self.assertEqual(hash(font_one), hash(font_one))

    def test_hash_different_object(self):
        font_one = self.getFont_glyphs()
        font_two = self.getFont_glyphs()
        self.assertNotEqual(hash(font_one), hash(font_two))

    def test_hash_same_object_variable_assignment(self):
        font_one = self.getFont_glyphs()
        a = font_one
        self.assertEqual(hash(font_one), hash(a))

    def test_hash_different_object_variable_assignment(self):
        font_one = self.getFont_glyphs()
        font_two = self.getFont_glyphs()
        a = font_one
        self.assertNotEqual(hash(font_two), hash(a))

    def test_hash_is_hasbable(self):
        font_one = self.getFont_glyphs()
        self.assertEqual(isinstance(font_one, collections.abc.Hashable), True)

    # --------
    # Equality
    # --------

    def test_object_equal_self(self):
        font_one = self.getFont_glyphs()
        self.assertEqual(font_one, font_one)

    def test_object_not_equal_other(self):
        font_one = self.getFont_glyphs()
        font_two = self.getFont_glyphs()
        self.assertNotEqual(font_one, font_two)

    def test_object_equal_self_variable_assignment(self):
        font_one = self.getFont_glyphs()
        a = font_one
        a.newGlyph("XYZ")
        self.assertEqual(font_one, a)

    def test_object_not_equal_other_variable_assignment(self):
        font_one = self.getFont_glyphs()
        font_two = self.getFont_glyphs()
        a = font_one
        self.assertNotEqual(font_two, a)

    # ---------
    # Selection
    # ---------

    # Font

    def test_selected_true(self):
        font = self.getFont_glyphs()
        try:
            font.selected = False
        except NotImplementedError:
            return
        font.selected = True
        self.assertEqual(font.selected, True)

    def test_selected_false(self):
        font = self.getFont_glyphs()
        try:
            font.selected = False
        except NotImplementedError:
            return
        self.assertEqual(font.selected, False)

    # Layers

    def test_selectedLayer_default(self):
        font = self.getFont_layers()
        try:
            font.defaultLayer.selected = False
        except NotImplementedError:
            return
        self.assertEqual(font.selectedLayers, ())

    def test_selectedLayer_setSubObject(self):
        font = self.getFont_layers()
        try:
            font.defaultLayer.selected = False
        except NotImplementedError:
            return
        layer1 = font.getLayer("layer A")
        layer2 = font.getLayer("layer B")
        layer1.selected = True
        layer2.selected = True
        self.assertEqual(font.selectedLayers, (layer1, layer2))

    def test_selectedLayer_setFilledList(self):
        font = self.getFont_layers()
        try:
            font.defaultLayer.selected = False
        except NotImplementedError:
            return
        layer3 = font.getLayer("layer C")
        layer4 = font.getLayer("layer D")
        font.selectedLayers = [layer3, layer4]
        self.assertEqual(font.selectedLayers, (layer3, layer4))

    def test_selectedLayer_setEmptyList(self):
        font = self.getFont_layers()
        try:
            font.defaultLayer.selected = False
        except NotImplementedError:
            return
        layer1 = font.getLayer("layer A")
        layer1.selected = True
        font.selectedLayers = []
        self.assertEqual(font.selectedLayers, ())

    # Layer names

    def test_selectedLayerNames_default(self):
        font = self.getFont_layers()
        try:
            font.defaultLayer.selected = False
        except NotImplementedError:
            return
        self.assertEqual(font.selectedLayers, ())

    def test_selectedLayerNames_setSubObject(self):
        font = self.getFont_layers()
        try:
            font.defaultLayer.selected = False
        except NotImplementedError:
            return
        layer1 = font.getLayer("layer A")
        layer2 = font.getLayer("layer B")
        layer1.selected = True
        layer2.selected = True
        self.assertEqual(tuple(sorted(font.selectedLayerNames)), ("layer A", "layer B"))

    def test_selectedLayerNames_setFilledList(self):
        font = self.getFont_layers()
        try:
            font.defaultLayer.selected = False
        except NotImplementedError:
            return
        font.selectedLayerNames = ["layer C", "layer D"]
        self.assertEqual(tuple(sorted(font.selectedLayerNames)), ("layer C", "layer D"))

    def test_selectedLayerNames_setEmptyList(self):
        font = self.getFont_layers()
        try:
            font.defaultLayer.selected = False
        except NotImplementedError:
            return
        layer1 = font.getLayer("layer A")
        layer1.selected = True
        font.selectedLayerNames = []
        self.assertEqual(font.selectedLayerNames, ())

    # Glyphs

    def test_selectedGlyphs_default(self):
        font = self.getFont_glyphs()
        try:
            font.defaultLayer.selected = False
        except NotImplementedError:
            return
        self.assertEqual(font.selectedGlyphs, ())

    def test_selectedGlyphs_setSubObject(self):
        font = self.getFont_glyphs()
        try:
            font.defaultLayer.selected = False
        except NotImplementedError:
            return
        glyph1 = font["A"]
        glyph2 = font["B"]
        glyph1.selected = True
        glyph2.selected = True
        self.assertEqual(
            tuple(sorted(font.selectedGlyphs, key=lambda glyph: glyph.name)),
            (glyph1, glyph2),
        )

    def test_selectedGlyphs_setFilledList(self):
        font = self.getFont_glyphs()
        try:
            font.defaultLayer.selected = False
        except NotImplementedError:
            return
        glyph3 = font["C"]
        glyph4 = font["D"]
        font.selectedGlyphs = [glyph3, glyph4]
        self.assertEqual(
            tuple(sorted(font.selectedGlyphs, key=lambda glyph: glyph.name)),
            (glyph3, glyph4),
        )

    def test_selectedGlyphs_setEmptyList(self):
        font = self.getFont_glyphs()
        try:
            font.defaultLayer.selected = False
        except NotImplementedError:
            return
        glyph1 = font["A"]
        glyph1.selected = True
        font.selectedGlyphs = []
        self.assertEqual(font.selectedGlyphs, ())

    # Glyph names

    def test_selectedGlyphNames_default(self):
        font = self.getFont_glyphs()
        try:
            font.defaultLayer.selected = False
        except NotImplementedError:
            return
        self.assertEqual(font.selectedGlyphs, ())

    def test_selectedGlyphNames_setSubObject(self):
        font = self.getFont_glyphs()
        try:
            font.defaultLayer.selected = False
        except NotImplementedError:
            return
        glyph1 = font["A"]
        glyph2 = font["B"]
        glyph1.selected = True
        glyph2.selected = True
        self.assertEqual(tuple(sorted(font.selectedGlyphNames)), ("A", "B"))

    def test_selectedGlyphNames_setFilledList(self):
        font = self.getFont_glyphs()
        try:
            font.defaultLayer.selected = False
        except NotImplementedError:
            return
        font.selectedGlyphNames = ["C", "D"]
        self.assertEqual(tuple(sorted(font.selectedGlyphNames)), ("C", "D"))

    def test_selectedGlyphNames_setEmptyList(self):
        font = self.getFont_glyphs()
        try:
            font.defaultLayer.selected = False
        except NotImplementedError:
            return
        glyph1 = font["A"]
        glyph1.selected = True
        font.selectedGlyphNames = []
        self.assertEqual(font.selectedGlyphNames, ())

    # Guidelines

    def test_selectedGuidelines_default(self):
        font = self.getFont_guidelines()
        guideline1 = font.guidelines[0]
        try:
            guideline1.selected = False
        except NotImplementedError:
            return
        self.assertEqual(font.selectedGuidelines, ())

    def test_selectedGuidelines_setSubObject(self):
        font = self.getFont_guidelines()
        guideline1 = font.guidelines[0]
        guideline2 = font.guidelines[1]
        try:
            guideline1.selected = False
        except NotImplementedError:
            return
        guideline2.selected = True
        self.assertEqual(font.selectedGuidelines, (guideline2,))

    def test_selectedGuidelines_setFilledList(self):
        font = self.getFont_guidelines()
        guideline1 = font.guidelines[0]
        guideline2 = font.guidelines[1]
        try:
            guideline1.selected = False
        except NotImplementedError:
            return
        font.selectedGuidelines = [guideline1, guideline2]
        self.assertEqual(font.selectedGuidelines, (guideline1, guideline2))

    def test_selectedGuidelines_setEmptyList(self):
        font = self.getFont_guidelines()
        guideline1 = font.guidelines[0]
        try:
            guideline1.selected = True
        except NotImplementedError:
            return
        font.selectedGuidelines = []
        self.assertEqual(font.selectedGuidelines, ())

    # ---------------
    # File Operations
    # ---------------

    # save

    def _saveFontPath(self, ext):
        root = tempfile.mkdtemp()
        return os.path.join(root, f"test{ext}")

    def _tearDownPath(self, path):
        if os.path.isdir(path):
            shutil.rmtree(path)
        elif os.path.isfile(path):
            os.remove(path)

    def _save(self, testCallback, **kwargs):
        path = self._saveFontPath(".ufo")
        font = self.getFont_glyphs()
        font.save(path, **kwargs)
        font.close()
        testCallback(path)
        self._tearDownPath(path)

    def test_save(self):
        def testCases(path):
            self.assertTrue(os.path.exists(path) and os.path.isdir(path))

        self._save(testCases)

    def test_save_formatVersion(self):
        from fontTools.ufoLib import UFOReader

        for version in [
            2,
            3,
        ]:  # fails on formatVersion 1 (but maybe we should not worry about it...)

            def testCases(path):
                reader = UFOReader(path)
                self.assertEqual(reader.formatVersion, version)

            self._save(testCases, formatVersion=version)

    def test_save_fileStructure(self):
        from fontTools.ufoLib import UFOReader, UFOFileStructure

        for fileStructure in [None, "package", "zip"]:

            def testCases(path):
                reader = UFOReader(path)
                expectedFileStructure = fileStructure
                if fileStructure is None:
                    expectedFileStructure = UFOFileStructure.PACKAGE
                else:
                    expectedFileStructure = UFOFileStructure(fileStructure)
                self.assertEqual(reader.fileStructure, expectedFileStructure)

            self._save(testCases, fileStructure=fileStructure)

    def test_save_path_none(self):
        font = self.getFont_glyphs()
        with self.assertRaises(OSError):
            font.save(path=None)

    # close

    def test_close(self):
        font, _ = self.objectGenerator("font")
        with patch.object(type(font), "_close") as mock_close:
            font.close(save=False)
            mock_close.assert_called_once()

    def test_save_and_close(self):
        font, _ = self.objectGenerator("font")
        with (
            patch.object(type(font), "save") as mock_save,
            patch.object(type(font), "_close") as mock_close,
        ):
            font.close(save=True)
            mock_close.assert_called_once()
            mock_save.assert_called_once()

    # copy

    def test_copy(self):
        font = self.getFont_glyphs()
        copy = font.copy()
        self.assertCountEqual(font.keys(), copy.keys())

        font = self.getFont_glyphs()
        font.defaultLayer.name = "hello"
        copy = font.copy()
        self.assertCountEqual(font.keys(), copy.keys())
        self.assertEqual(font.defaultLayerName, copy.defaultLayerName)

        font = self.getFont_guidelines()
        copy = font.copy()
        self.assertEqual(copy.selectedGuidelines, font.selectedGuidelines)

    def test_copyData_source_layer_not_in_layerOrder(self):
        source = self.getFont_layers()
        font, _ = self.objectGenerator("font")
        font.copyData(source)
        self.assertIn("layer A", font.layerOrder)

    # generate

    def test_generateFormatToExtension(self):
        font, _ = self.objectGenerator("font")
        cases = [
            ("macttf", ".ttf"),
            ("macttdfont", ".dfont"),
            ("otfcff", ".otf"),
            ("otfttf", ".ttf"),
            ("ufo1", ".ufo"),
            ("ufo2", ".ufo"),
            ("ufo3", ".ufo"),
            ("unixascii", ".pfa"),
        ]
        for fmt, expected_ext in cases:
            with self.subTest(format=fmt):
                result = font.generateFormatToExtension(fmt, fallbackFormat=".fallback")
                self.assertEqual(result, expected_ext)

    def test_generate_format_none(self):
        font = self.getFont_glyphs()
        with self.assertRaises(ValueError):
            font.generate(format=None)

    def test_generate_invalid_format_type(self):
        font = self.getFont_glyphs()
        with self.assertRaises(TypeError):
            font.generate(format=0)

    def test_generate_valid_environmentOption(self):
        import warnings

        font = self.getFont_glyphs()
        with (
            patch.object(
                type(font), "_isValidGenerateEnvironmentOption"
            ) as mockValidate,
            patch.object(type(font), "_generate") as mockGenerate,
        ):
            mockValidate.return_value = True
            with warnings.catch_warnings():
                warnings.simplefilter("error", UserWarning)
                font.generate("otfcff", path="/mock/path.otf", valid_opt="value")

            mockGenerate.assert_called_once_with(
                format="otfcff",
                path="/mock/path.otf",
                environmentOptions={"valid_opt": "value"},
            )

    def test_generate_invalid_environmentOption(self):
        font = self.getFont_glyphs()
        with (
            patch.object(
                type(font), "_isValidGenerateEnvironmentOption"
            ) as mockValidate,
            patch.object(type(font), "_generate") as mockGenerate,
        ):
            mockValidate.return_value = False
            with self.assertWarns(UserWarning):
                font.generate("otfcff", path="/mock/path.otf", invalid_opt="value")

            mockGenerate.assert_called_once_with(
                format="otfcff",
                path="/mock/path.otf",
                environmentOptions={"invalid_opt": "value"},
            )

    def test_generate_path_none_and_self_path_none(self):
        font = self.getFont_glyphs()
        with patch.object(
            type(font), "_isValidGenerateEnvironmentOption"
        ) as mockValidate:
            mockValidate.return_value = True
            with self.assertRaises(OSError):
                font.generate("otfcff", path=None)

    def test_generate_path_none(self):
        font = self.getFont_glyphs()
        with (
            patch.object(type(font), "path", new_callable=PropertyMock) as mockPath,
            patch.object(type(font), "_generate") as mockGenerate,
        ):
            mockPath.return_value = "/src/myfont.otf"
            font.generate("otfcff", path=None)
            expectedPath = os.path.abspath("/src/myfont.otf")
            mockGenerate.assert_called_once()
            actualPath = mockGenerate.call_args[1]["path"]
            self.assertEqual(os.path.abspath(actualPath), os.path.abspath(expectedPath))

    def test_generate_path_is_dir_and_self_path_none(self):
        font = self.getFont_glyphs()
        with (
            patch.object(type(font), "path", new_callable=PropertyMock) as mockPath,
            patch.object(os.path, "isdir") as mockIsDir,
        ):
            mockPath.return_value = None
            mockIsDir.return_value = True
            with self.assertRaises(OSError):
                font.generate("otfcff", path="/output/dir")

    def test_generate_path_is_dir(self):
        font = self.getFont_glyphs()
        with (
            patch.object(type(font), "path", new_callable=PropertyMock) as mockPath,
            patch.object(os.path, "isdir") as mockIsDir,
            patch.object(type(font), "_generate") as mockGenerate,
        ):
            mockPath.return_value = "/src/myfont.ufo"
            mockIsDir.return_value = True
            font.generate("otfcff", path="/output/dir")
            expectedPath = os.path.abspath(os.path.join("/output/dir", "myfont.otf"))
            actualPath = mockGenerate.call_args[1]["path"]
            self.assertEqual(os.path.abspath(actualPath), os.path.abspath(expectedPath))

    # -----------------
    # Global Operations
    # -----------------

    def test_round(self):
        font, _ = self.objectGenerator("font")
        font.info.xHeight = 450.6
        font.kerning[("A", "V")] = -20.3
        guideline = font.appendGuideline((100.4, 200.7), 0)
        defaultLayer = font.defaultLayer
        glyph = defaultLayer.newGlyph("A")
        glyph.width = 600.8
        font.round()
        self.assertEqual(font.info.xHeight, 451)
        self.assertEqual(font.kerning[("A", "V")], -20)
        self.assertEqual(guideline.position, (100, 201))
        self.assertEqual(glyph.width, 601)

    # -------------
    # Interpolation
    # -------------

    def test_interpolate(self):
        dstFont, _ = self.objectGenerator("font")
        minFont, _ = self.objectGenerator("font")
        maxFont, _ = self.objectGenerator("font")
        minFont.info.ascender = 100
        minFont.kerning[("A", "V")] = -20
        minLayer = minFont.getLayer(minFont.defaultLayer.name)
        minGlyph = minLayer.newGlyph("A")
        minGlyph.width = 100
        maxFont.info.ascender = 200
        maxFont.kerning[("A", "V")] = -120
        maxLayer = maxFont.getLayer(maxFont.defaultLayer.name)
        maxGlyph = maxLayer.newGlyph("A")
        maxGlyph.width = 200
        dstFont.interpolate(0.5, minFont, maxFont, round=True)
        self.assertEqual(dstFont.info.ascender, 150)
        self.assertEqual(dstFont.kerning[("A", "V")], -70)
        self.assertIn("A", dstFont)
        self.assertEqual(dstFont["A"].width, 150)

    def test_interpolate_invalid_types(self):
        dstFont, _ = self.objectGenerator("font")
        srcFont, _ = self.objectGenerator("font")
        maxFont, _ = self.objectGenerator("font")
        testCases = [
            (0.5, srcFont, "ivalidFontObject"),
            (0.5, "ivalidFontObject", srcFont),
        ]
        for factor, minFont, maxFont in testCases:
            with self.assertRaises(TypeError):
                dstFont.interpolate(factor, minFont, maxFont)

    def test_interpolate_asymmetric_layers(self):
        dstFont, _ = self.objectGenerator("font")
        minFont, _ = self.objectGenerator("font")
        maxFont, _ = self.objectGenerator("font")
        minFont.newLayer("testLayer")
        dstFont.interpolate(0.5, minFont, maxFont)
        self.assertNotIn("testLayer", dstFont.layerOrder)
        self.assertIn(DEFAULT_LAYER_NAME, dstFont.layerOrder)

    @patch("fontParts.base.font.ufoLib.DEFAULT_LAYER_NAME", "invalid.default.layer")
    def test_interpolate_missing_default_layer(self):
        dstFont, _ = self.objectGenerator("font")
        minFont, _ = self.objectGenerator("font")
        maxFont, _ = self.objectGenerator("font")
        dstFont.newLayer("testLayer")
        dstFont.interpolate(0.5, minFont, maxFont)
        self.assertEqual(dstFont.defaultLayer.name, dstFont.layerOrder[0])

    def test_interpolate_global_guidelines(self):
        interpolated_font, _ = self.objectGenerator("font")
        font_min, _ = self.objectGenerator("font")
        font_min.appendGuideline(position=(0, 0), angle=0)
        font_max, _ = self.objectGenerator("font")
        font_max.appendGuideline(position=(200, 200), angle=0)
        interpolated_font.info.interpolate(
            0.5, font_min.info, font_max.info, round=True
        )
        self.assertEqual(len(interpolated_font.guidelines), 1)
        self.assertEqual(interpolated_font.guidelines[0].position, (100, 100))

    def test_isCompatible_guideline_counts_differ(self):
        font1 = self.getFont_guidelines()
        font2 = self.getFont_guidelines()
        font2.appendGuideline((10, 10))
        compatible, report = font1.isCompatible(font2)
        self.assertTrue(compatible)
        self.assertTrue(report.warning)
        self.assertTrue(report.guidelineCountDifference)

    def test_isCompatible_guidelines_differ(self):
        font1 = self.getFont_guidelines()
        font2 = self.getFont_guidelines()
        font2.guidelines[0].name = "otherGuideline"
        compatible, report = font1.isCompatible(font2)
        self.assertTrue(compatible)
        self.assertTrue(report.warning)
        self.assertTrue(report.guidelinesMissingInFont1)

    def test_isCompatible_layer_counts_differ(self):
        font1 = self.getFont_glyphs()
        font2 = self.getFont_glyphs()
        font2.newLayer("testLayer")
        compatible, report = font1.isCompatible(font2)
        self.assertTrue(compatible)
        self.assertTrue(report.warning)
        self.assertTrue(report.layerCountDifference)

    def test_isCompatible_layers_differ(self):
        font1 = self.getFont_glyphs()
        font2 = self.getFont_glyphs()
        font2.layers[0].name = "otherLayer"
        compatible, report = font1.isCompatible(font2)
        self.assertTrue(compatible)
        self.assertTrue(report.warning)
        self.assertTrue(report.layersMissingFromFont2)

    def test_isCompatible_layer_fatal_only(self):
        font1 = self.getFont_glyphs()
        font2 = self.getFont_glyphs()
        glyph1 = font1.newGlyph("A")
        glyph2 = font2.newGlyph("A")
        pen1 = glyph1.getPointPen()
        pen1.beginPath()
        pen1.addPoint((0, 0), segmentType="line")
        pen1.endPath()
        pen2 = glyph2.getPointPen()
        pen2.beginPath()
        pen2.addPoint((0, 0), segmentType="line")
        pen2.addPoint((100, 100), segmentType="line")
        pen2.endPath()
        compatible, report = font1.isCompatible(font2)
        self.assertFalse(compatible)
        self.assertTrue(report.fatal)
        self.assertFalse(report.warning)

    def test_isCompatible_layer_warning_only(self):
        font1 = self.getFont_glyphs()
        font2 = self.getFont_glyphs()
        glyph1 = font1.newGlyph("A")
        glyph2 = font2.newGlyph("A")
        glyph1.appendAnchor("top", (50, 50))
        glyph2.appendAnchor("bottom", (50, 50))
        compatible, report = font1.isCompatible(font2)
        self.assertTrue(compatible)
        self.assertFalse(report.fatal)
        self.assertTrue(report.warning)

    # -------
    # Mapping
    # -------

    def getFont_components(self):
        font = self.getFont_glyphs()
        composite = font.newGlyph("compositeGlyph")
        for baseName in "ABCD":
            composite.appendComponent(baseName)
        return font

    def test_getReverseComponentMapping(self):
        font = self.getFont_components()
        fontMapping = font.getReverseComponentMapping()
        layerMapping = font.defaultLayer.getReverseComponentMapping()
        self.assertEqual(fontMapping, layerMapping)
        self.assertIn("A", fontMapping)

    def test_characterMapping(self):
        font = self.getFont_glyphs()
        for i, glyph in enumerate(font):
            glyph.unicode = i % 2
        fontMapping = font.getCharacterMapping()
        layerMapping = font.defaultLayer.getCharacterMapping()
        self.assertEqual(fontMapping, layerMapping)
        self.assertIn(1, fontMapping)



from fontPens.recordingPointPen import RecordingPointPen


class LegacyPointPen(RecordingPointPen):
    """
    A point pen that accepts only the original
    arguments in the various methods.
    """

    def beginPath(self):
        super().beginPath()

    def endPath(self):
        super().endPath()

    def addPoint(self, pt, segmentType=None, smooth=False, name=None):
        super().addPoint(pt, segmentType=segmentType, smooth=smooth, name=name)

    def addComponent(self, baseGlyphName, transformation):
        super().addComponent(baseGlyphName, transformation)



import collections
import unittest

from fontParts.base import FontPartsError


class TestComponent(unittest.TestCase):
    def getComponent_generic(self):
        layer, _ = self.objectGenerator("layer")
        glyph = layer.newGlyph("A")
        pen = glyph.getPen()
        pen.moveTo((0, 0))
        pen.lineTo((0, 100))
        pen.lineTo((100, 100))
        pen.lineTo((100, 0))
        pen.closePath()
        for i, point in enumerate(glyph[0].points):
            point.name = f"point {i}"
        glyph = layer.newGlyph("B")
        component = glyph.appendComponent("A")
        component.transformation = (1, 0, 0, 1, 0, 0)
        return component

    # ----
    # repr
    # ----

    def test_reprContents(self):
        component = self.getComponent_generic()
        value = component._reprContents()
        self.assertIsInstance(value, list)
        for i in value:
            self.assertIsInstance(i, str)

    def test_reprContents_noGlyph(self):
        component, _ = self.objectGenerator("component")
        value = component._reprContents()
        self.assertIsInstance(value, list)
        for i in value:
            self.assertIsInstance(i, str)

    # -------
    # Parents
    # -------

    def test_get_parent_font(self):
        font, _ = self.objectGenerator("font")
        layer = font.newLayer("L")
        glyph = layer.newGlyph("X")
        component = glyph.appendComponent("A")
        self.assertIsNotNone(component.font)
        self.assertEqual(component.font, font)

    def test_get_parent_noFont(self):
        layer, _ = self.objectGenerator("layer")
        glyph = layer.newGlyph("X")
        component = glyph.appendComponent("A")
        self.assertIsNone(component.font)

    def test_get_parent_layer(self):
        layer, _ = self.objectGenerator("layer")
        glyph = layer.newGlyph("X")
        component = glyph.appendComponent("A")
        self.assertIsNotNone(component.layer)
        self.assertEqual(component.layer, layer)

    def test_get_parent_noLayer(self):
        glyph, _ = self.objectGenerator("glyph")
        component = glyph.appendComponent("A")
        self.assertIsNone(component.font)
        self.assertIsNone(component.layer)

    def test_get_parent_glyph(self):
        glyph, _ = self.objectGenerator("glyph")
        component = glyph.appendComponent("A")
        self.assertIsNotNone(component.glyph)
        self.assertEqual(component.glyph, glyph)

    def test_get_parent_noGlyph(self):
        component, _ = self.objectGenerator("component")
        self.assertIsNone(component.font)
        self.assertIsNone(component.layer)
        self.assertIsNone(component.glyph)

    def test_set_parent_glyph(self):
        glyph, _ = self.objectGenerator("glyph")
        component, _ = self.objectGenerator("component")
        component.glyph = glyph
        self.assertIsNotNone(component.glyph)
        self.assertEqual(component.glyph, glyph)

    def test_set_parent_glyph_none(self):
        component, _ = self.objectGenerator("component")
        component.glyph = None
        self.assertIsNone(component.glyph)

    def test_set_parent_glyph_exists(self):
        glyph, _ = self.objectGenerator("glyph")
        otherGlyph, _ = self.objectGenerator("glyph")
        component = glyph.appendComponent("A")
        with self.assertRaises(AssertionError):
            component.glyph = otherGlyph

    # ----------
    # Attributes
    # ----------

    # baseGlyph

    def test_baseGlyph_generic(self):
        component = self.getComponent_generic()
        self.assertEqual(component.baseGlyph, "A")

    def test_baseGlyph_valid_set(self):
        component = self.getComponent_generic()
        component.baseGlyph = "B"
        self.assertEqual(component.baseGlyph, "B")

    def test_baseGlyph_invalid_set_none(self):
        component = self.getComponent_generic()
        with self.assertRaises(TypeError):
            component.baseGlyph = None

    def test_baseGlyph_invalid_set_empty_string(self):
        component = self.getComponent_generic()
        with self.assertRaises(ValueError):
            component.baseGlyph = ""

    def test_baseGlyph_invalid_set_int(self):
        component = self.getComponent_generic()
        with self.assertRaises(TypeError):
            component.baseGlyph = 123

    # transformation

    def test_transformation_generic(self):
        component = self.getComponent_generic()
        self.assertEqual(component.transformation, (1, 0, 0, 1, 0, 0))

    def test_transformation_valid_set_positive(self):
        component = self.getComponent_generic()
        component.transformation = (1, 2, 3, 4, 5, 6)
        self.assertEqual(component.transformation, (1, 2, 3, 4, 5, 6))

    def test_transformation_valid_set_negative(self):
        component = self.getComponent_generic()
        component.transformation = (-1, -2, -3, -4, -5, -6)
        self.assertEqual(component.transformation, (-1, -2, -3, -4, -5, -6))

    def test_transformation_invalid_set_member(self):
        component = self.getComponent_generic()
        with self.assertRaises(TypeError):
            component.transformation = (1, 0, 0, 1, 0, "0")

    def test_transformation_invalid_set_none(self):
        component = self.getComponent_generic()
        with self.assertRaises(TypeError):
            component.transformation = None

    # offset

    def test_offset_generic(self):
        component = self.getComponent_generic()
        self.assertEqual(component.offset, (0, 0))

    def test_offset_valid_set_zero(self):
        component = self.getComponent_generic()
        component.offset = (0, 0)
        self.assertEqual(component.offset, (0, 0))

    def test_offset_valid_set_positive_positive(self):
        component = self.getComponent_generic()
        component.offset = (1, 2)
        self.assertEqual(component.offset, (1, 2))

    def test_offset_valid_set_negative_positive(self):
        component = self.getComponent_generic()
        component.offset = (-1, 2)
        self.assertEqual(component.offset, (-1, 2))

    def test_offset_valid_set_positive_negative(self):
        component = self.getComponent_generic()
        component.offset = (1, -2)
        self.assertEqual(component.offset, (1, -2))

    def test_offset_valid_set_negative_negative(self):
        component = self.getComponent_generic()
        component.offset = (-1, -2)
        self.assertEqual(component.offset, (-1, -2))

    def test_offset_invalid_set_real_bogus(self):
        component = self.getComponent_generic()
        with self.assertRaises(TypeError):
            component.offset = (1, "2")

    def test_offset_invalid_set_bogus_real(self):
        component = self.getComponent_generic()
        with self.assertRaises(TypeError):
            component.offset = ("1", 2)

    def test_offset_invalid_set_int(self):
        component = self.getComponent_generic()
        with self.assertRaises(TypeError):
            component.offset = 1

    def test_offset_invalid_set_none(self):
        component = self.getComponent_generic()
        with self.assertRaises(TypeError):
            component.offset = None

    # scale

    def test_scale_generic(self):
        component = self.getComponent_generic()
        self.assertEqual(component.scale, (1, 1))

    def test_scale_valid_set_zero(self):
        component = self.getComponent_generic()
        component.scale = (0, 0)
        self.assertEqual(component.scale, (0, 0))

    def test_scale_valid_set_positive_positive(self):
        component = self.getComponent_generic()
        component.scale = (1, 2)
        self.assertEqual(component.scale, (1, 2))

    def test_scale_valid_set_negative_positive(self):
        component = self.getComponent_generic()
        component.scale = (-1, 2)
        self.assertEqual(component.scale, (-1, 2))

    def test_scale_valid_set_positive_negative(self):
        component = self.getComponent_generic()
        component.scale = (1, -2)
        self.assertEqual(component.scale, (1, -2))

    def test_scale_valid_set_negative_negative(self):
        component = self.getComponent_generic()
        component.scale = (-1, -2)
        self.assertEqual(component.scale, (-1, -2))

    def test_scale_invalid_set_real_bogus(self):
        component = self.getComponent_generic()
        with self.assertRaises(TypeError):
            component.scale = (1, "2")

    def test_scale_invalid_set_bogus_real(self):
        component = self.getComponent_generic()
        with self.assertRaises(TypeError):
            component.scale = ("1", 2)

    def test_scale_invalid_set_int(self):
        component = self.getComponent_generic()
        with self.assertRaises(TypeError):
            component.scale = 1

    def test_scale_invalid_set_none(self):
        component = self.getComponent_generic()
        with self.assertRaises(TypeError):
            component.scale = None

    # --------------
    # Identification
    # --------------

    # index

    def getComponent_index(self):
        glyph, _ = self.objectGenerator("glyph")
        glyph.appendComponent("A")
        glyph.appendComponent("B")
        glyph.appendComponent("C")
        return glyph

    def test_get_index_noParent(self):
        component, _ = self.objectGenerator("component")
        self.assertIsNone(component.index)

    def test_get_index(self):
        glyph = self.getComponent_index()
        for i, component in enumerate(glyph.components):
            self.assertEqual(component.index, i)

    def test_set_index_noParent(self):
        component, _ = self.objectGenerator("component")
        with self.assertRaises(FontPartsError):
            component.index = 1

    def test_set_index_positive(self):
        glyph = self.getComponent_index()
        component = glyph.components[0]
        component.index = 2
        self.assertEqual([c.baseGlyph for c in glyph.components], ["B", "A", "C"])

    def test_set_index_pastLength(self):
        glyph = self.getComponent_index()
        component = glyph.components[0]
        component.index = 20
        self.assertEqual([c.baseGlyph for c in glyph.components], ["B", "C", "A"])

    def test_set_index_negative(self):
        glyph = self.getComponent_index()
        component = glyph.components[1]
        component.index = -1
        self.assertEqual([c.baseGlyph for c in glyph.components], ["B", "A", "C"])

    # identifier

    def test_identifier_get_none(self):
        component = self.getComponent_generic()
        self.assertIsNone(component.identifier)

    def test_identifier_generated_type(self):
        component = self.getComponent_generic()
        component.getIdentifier()
        self.assertIsInstance(component.identifier, str)

    def test_identifier_consistency(self):
        component = self.getComponent_generic()
        component.getIdentifier()
        # get: twice to test consistency
        self.assertEqual(component.identifier, component.identifier)

    def test_identifier_cannot_set(self):
        # identifier is a read-only property
        component = self.getComponent_generic()
        with self.assertRaises(FontPartsError):
            component.identifier = "ABC"

    # ----
    # Copy
    # ----

    def getComponent_copy(self):
        component = self.getComponent_generic()
        component.transformation = (1, 2, 3, 4, 5, 6)
        return component

    def test_copy_seperate_objects(self):
        component = self.getComponent_copy()
        copied = component.copy()
        self.assertIsNot(component, copied)

    def test_copy_same_baseGlyph(self):
        component = self.getComponent_copy()
        copied = component.copy()
        self.assertEqual(component.baseGlyph, copied.baseGlyph)

    def test_copy_same_transformation(self):
        component = self.getComponent_copy()
        copied = component.copy()
        self.assertEqual(component.transformation, copied.transformation)

    def test_copy_same_offset(self):
        component = self.getComponent_copy()
        copied = component.copy()
        self.assertEqual(component.offset, copied.offset)

    def test_copy_same_scale(self):
        component = self.getComponent_copy()
        copied = component.copy()
        self.assertEqual(component.scale, copied.scale)

    def test_copy_not_identifier(self):
        component = self.getComponent_copy()
        component.getIdentifier()
        copied = component.copy()
        self.assertNotEqual(component.identifier, copied.identifier)

    def test_copy_generated_identifier_different(self):
        component = self.getComponent_copy()
        copied = component.copy()
        component.getIdentifier()
        copied.getIdentifier()
        self.assertNotEqual(component.identifier, copied.identifier)

    # ----
    # Pens
    # ----

    # draw

    def test_draw(self):
        from fontTools.pens.recordingPen import RecordingPen

        component = self.getComponent_generic()
        component.transformation = (1, 2, 3, 4, 5, 6)
        pen = RecordingPen()
        component.draw(pen)
        expected = [("addComponent", ("A", (1.0, 2.0, 3.0, 4.0, 5.0, 6.0)))]
        self.assertEqual(pen.value, expected)

    # drawPoints

    def test_drawPoints(self):
        from fontPens.recordingPointPen import RecordingPointPen

        component = self.getComponent_generic()
        component.transformation = (1, 2, 3, 4, 5, 6)
        identifier = component.getIdentifier()
        pointPen = RecordingPointPen()
        component.drawPoints(pointPen)
        expected = [
            (
                "addComponent",
                ("A", (1.0, 2.0, 3.0, 4.0, 5.0, 6.0)),
                {"identifier": identifier},
            )
        ]
        self.assertEqual(pointPen.value, expected)

    def test_drawPoints_legacy(self):
        from .legacyPointPen import LegacyPointPen

        component = self.getComponent_generic()
        component.transformation = (1, 2, 3, 4, 5, 6)
        component.getIdentifier()
        pointPen = LegacyPointPen()
        component.drawPoints(pointPen)
        expected = [("addComponent", ("A", (1.0, 2.0, 3.0, 4.0, 5.0, 6.0)), {})]
        self.assertEqual(pointPen.value, expected)

    # --------------
    # Transformation
    # --------------

    def getComponent_transform(self):
        component = self.getComponent_generic()
        component.transformation = (1, 2, 3, 4, 5, 6)
        return component

    # transformBy

    def test_transformBy_valid_no_origin(self):
        component = self.getComponent_transform()
        component.transformBy((2, 0, 0, 3, -3, 2))
        self.assertEqual(component.transformation, (2.0, 6.0, 6.0, 12.0, 7.0, 20.0))

    def test_transformBy_valid_origin(self):
        component = self.getComponent_transform()
        component.transformBy((2, 0, 0, 2, 0, 0), origin=(1, 2))
        self.assertEqual(component.transformation, (2.0, 4.0, 6.0, 8.0, 9.0, 10.0))

    def test_transformBy_invalid_one_string_value(self):
        component = self.getComponent_transform()
        with self.assertRaises(TypeError):
            component.transformBy((1, 0, 0, 1, 0, "0"))

    def test_transformBy_invalid_all_string_values(self):
        component = self.getComponent_transform()
        with self.assertRaises(TypeError):
            component.transformBy("1, 0, 0, 1, 0, 0")

    def test_transformBy_invalid_int_value(self):
        component = self.getComponent_transform()
        with self.assertRaises(TypeError):
            component.transformBy(123)

    # moveBy

    def test_moveBy_valid(self):
        component = self.getComponent_transform()
        component.moveBy((-1, 2))
        self.assertEqual(component.transformation, (1.0, 2.0, 3.0, 4.0, 4.0, 8.0))

    def test_moveBy_invalid_one_string_value(self):
        component = self.getComponent_transform()
        with self.assertRaises(TypeError):
            component.moveBy((-1, "2"))

    def test_moveBy_invalid_all_strings_value(self):
        component = self.getComponent_transform()
        with self.assertRaises(TypeError):
            component.moveBy("-1, 2")

    def test_moveBy_invalid_int_value(self):
        component = self.getComponent_transform()
        with self.assertRaises(TypeError):
            component.moveBy(1)

    # scaleBy

    def test_scaleBy_valid_one_value_no_origin(self):
        component = self.getComponent_transform()
        component.scaleBy(-2)
        self.assertEqual(
            component.transformation, (-2.0, -4.0, -6.0, -8.0, -10.0, -12.0)
        )

    def test_scaleBy_valid_two_values_no_origin(self):
        component = self.getComponent_transform()
        component.scaleBy((-2, 3))
        self.assertEqual(component.transformation, (-2.0, 6.0, -6.0, 12.0, -10.0, 18.0))

    def test_scaleBy_valid_two_values_origin(self):
        component = self.getComponent_transform()
        component.scaleBy((-2, 3), origin=(1, 2))
        self.assertEqual(component.transformation, (-2.0, 6.0, -6.0, 12.0, -7.0, 14.0))

    def test_scaleBy_invalid_one_string_value(self):
        component = self.getComponent_transform()
        with self.assertRaises(TypeError):
            component.scaleBy((-1, "2"))

    def test_scaleBy_invalid_two_string_values(self):
        component = self.getComponent_transform()
        with self.assertRaises(TypeError):
            component.scaleBy("-1, 2")

    def test_scaleBy_invalid_tuple_too_many_values(self):
        component = self.getComponent_transform()
        with self.assertRaises(ValueError):
            component.scaleBy((-1, 2, -3))

    # # rotateBy

    def test_rotateBy_valid_no_origin(self):
        component = self.getComponent_transform()
        component.rotateBy(45)
        self.assertEqual(
            [round(i, 3) for i in component.transformation],
            [-0.707, 2.121, -0.707, 4.95, -0.707, 7.778],
        )

    def test_rotateBy_valid_origin(self):
        component = self.getComponent_transform()
        component.rotateBy(45, origin=(1, 2))
        self.assertEqual(
            [round(i, 3) for i in component.transformation],
            [-0.707, 2.121, -0.707, 4.95, 1.0, 7.657],
        )

    def test_rotateBy_invalid_string_value(self):
        component = self.getComponent_transform()
        with self.assertRaises(TypeError):
            component.rotateBy("45")

    def test_rotateBy_invalid_too_large_value_positive(self):
        component = self.getComponent_transform()
        with self.assertRaises(ValueError):
            component.rotateBy(361)

    def test_rotateBy_invalid_too_large_value_negative(self):
        component = self.getComponent_transform()
        with self.assertRaises(ValueError):
            component.rotateBy(-361)

    # skewBy

    def test_skewBy_valid_no_origin_one_value(self):
        component = self.getComponent_transform()
        component.skewBy(100)
        self.assertEqual(
            [round(i, 3) for i in component.transformation],
            [-10.343, 2.0, -19.685, 4.0, -29.028, 6.0],
        )

    def test_skewBy_valid_no_origin_two_values(self):
        component = self.getComponent_transform()
        component.skewBy((100, 200))
        self.assertEqual(
            [round(i, 3) for i in component.transformation],
            [-10.343, 2.364, -19.685, 5.092, -29.028, 7.82],
        )

    def test_skewBy_valid_origin_one_value(self):
        component = self.getComponent_transform()
        component.skewBy(100, origin=(1, 2))
        self.assertEqual(
            [round(i, 3) for i in component.transformation],
            [-10.343, 2.0, -19.685, 4.0, -17.685, 6.0],
        )

    def test_skewBy_valid_origin_two_values(self):
        component = self.getComponent_transform()
        component.skewBy((100, 200), origin=(1, 2))
        self.assertEqual(
            [round(i, 3) for i in component.transformation],
            [-10.343, 2.364, -19.685, 5.092, -17.685, 7.456],
        )

    # -------------
    # Normalization
    # -------------

    # round

    def test_round(self):
        component = self.getComponent_generic()
        component.transformation = (1.2, 2.2, 3.3, 4.4, 5.1, 6.6)
        component.round()
        self.assertEqual(component.transformation, (1.2, 2.2, 3.3, 4.4, 5, 7))

    # decompose

    def test_decompose_noParent(self):
        component, _ = self.objectGenerator("component")
        with self.assertRaises(FontPartsError):
            component.decompose()

    def test_decompose_digest(self):
        from fontPens.digestPointPen import DigestPointPen

        component = self.getComponent_generic()
        glyph = component.glyph
        glyph.layer[component.baseGlyph]
        component.decompose()
        pointPen = DigestPointPen()
        glyph.drawPoints(pointPen)
        expected = (
            "beginPath",
            ((0, 0), "line", False, "point 0"),
            ((0, 100), "line", False, "point 1"),
            ((100, 100), "line", False, "point 2"),
            ((100, 0), "line", False, "point 3"),
            "endPath",
        )
        self.assertEqual(pointPen.getDigest(), expected)

    def test_decompose_identifiers(self):
        component = self.getComponent_generic()
        glyph = component.glyph
        baseGlyph = glyph.layer[component.baseGlyph]
        baseGlyph[0].getIdentifier()
        for point in baseGlyph[0].points:
            point.getIdentifier()
        component.decompose()
        self.assertEqual(
            [point.identifier for point in glyph[0].points],
            [point.identifier for point in baseGlyph[0].points],
        )
        self.assertEqual(glyph[0].identifier, baseGlyph[0].identifier)

    def test_decompose_transformation(self):
        from fontPens.digestPointPen import DigestPointPen

        component = self.getComponent_generic()
        component.scale = (2, 2)
        glyph = component.glyph
        glyph.layer[component.baseGlyph]
        component.decompose()
        pointPen = DigestPointPen()
        glyph.drawPoints(pointPen)
        expected = (
            "beginPath",
            ((0, 0), "line", False, "point 0"),
            ((0, 200), "line", False, "point 1"),
            ((200, 200), "line", False, "point 2"),
            ((200, 0), "line", False, "point 3"),
            "endPath",
        )
        self.assertEqual(pointPen.getDigest(), expected)

    # ------------
    # Data Queries
    # ------------

    # bounds

    def test_bounds_get(self):
        component = self.getComponent_generic()
        self.assertEqual(component.bounds, (0, 0, 100, 100))

    def test_bounds_none(self):
        component = self.getComponent_generic()
        layer = component.layer
        baseGlyph = layer[component.baseGlyph]
        baseGlyph.clear()
        self.assertIsNone(component.bounds)

    def test_bounds_on_move(self):
        component = self.getComponent_generic()
        component.moveBy((0.1, -0.1))
        self.assertEqual(component.bounds, (0.1, -0.1, 100.1, 99.9))

    def test_bounds_on_scale(self):
        component = self.getComponent_generic()
        component.scaleBy((2, 0.5))
        self.assertEqual(component.bounds, (0, 0, 200, 50))

    def test_bounds_invalid_set(self):
        component = self.getComponent_generic()
        with self.assertRaises(FontPartsError):
            component.bounds = (0, 0, 100, 100)

    # pointInside

    def test_pointInside_true(self):
        component = self.getComponent_generic()
        self.assertEqual(component.pointInside((1, 1)), True)

    def test_pointInside_false(self):
        component = self.getComponent_generic()
        self.assertEqual(component.pointInside((-1, -1)), False)

    # ----
    # Hash
    # ----

    def test_hash_object_self(self):
        component_one = self.getComponent_generic()
        self.assertEqual(hash(component_one), hash(component_one))

    def test_hash_object_other(self):
        component_one = self.getComponent_generic()
        component_two = self.getComponent_generic()
        self.assertNotEqual(hash(component_one), hash(component_two))

    def test_hash_object_self_variable_assignment(self):
        component_one = self.getComponent_generic()
        a = component_one
        self.assertEqual(hash(component_one), hash(a))

    def test_hash_object_other_variable_assignment(self):
        component_one = self.getComponent_generic()
        component_two = self.getComponent_generic()
        a = component_one
        self.assertNotEqual(hash(component_two), hash(a))

    def test_is_hashable(self):
        component_one = self.getComponent_generic()
        self.assertTrue(isinstance(component_one, collections.abc.Hashable))

    # --------
    # Equality
    # --------

    def test_object_equal_self(self):
        component_one = self.getComponent_generic()
        self.assertEqual(component_one, component_one)

    def test_object_not_equal_other(self):
        component_one = self.getComponent_generic()
        component_two = self.getComponent_generic()
        self.assertNotEqual(component_one, component_two)

    def test_object_equal_assigned_variable(self):
        component_one = self.getComponent_generic()
        a = component_one
        a.baseGlyph = "C"
        self.assertEqual(component_one, a)

    def test_object_not_equal_assigned_variable_other(self):
        component_one = self.getComponent_generic()
        component_two = self.getComponent_generic()
        a = component_one
        self.assertNotEqual(component_two, a)

    # ---------
    # Selection
    # ---------

    def test_selected_true(self):
        component = self.getComponent_generic()
        try:
            component.selected = False
        except NotImplementedError:
            return
        component.selected = True
        self.assertEqual(component.selected, True)

    def test_selected_false(self):
        component = self.getComponent_generic()
        try:
            component.selected = False
        except NotImplementedError:
            return
        self.assertEqual(component.selected, False)



# pylint: disable=C0103, C0114
from __future__ import annotations
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, NoReturn, TypeVar
from collections.abc import Callable, Iterable, Iterator
from copy import deepcopy
import math
from collections.abc import MutableMapping

from fontTools.misc import transform
from fontParts.base.errors import FontPartsError
from fontParts.base import normalizers
from fontParts.base.annotations import (
    InterpolationFactorLike,
    SkewAngleLike,
    SkewAnglePair,
    ScaleFactorLike,
    ScaleFactorPair,
    AffineTransformationLike,
    Coordinate,
    CoordinateLike,
    IntFloatType,
    InterpolatableType,
)

if TYPE_CHECKING:
    from collections.abc import ItemsView

BaseObjectType = TypeVar("BaseObjectType", bound="BaseObject")
KeyType = TypeVar("KeyType")
ValueType = TypeVar("ValueType")

# -------
# Helpers
# -------


class dynamicProperty:
    """Represent a property for simplified subclassing.

    This class implements functionality that is very similar to Python's built
    in :class:`property`, but allows the getter and setter methods to be
    automatically resolved by name, making it possible to partially override
    the property in a subclass without having to re-register both the getter
    and setter methods.

    :param name: The base name of the property, used to resolve the getter and
        setter method names.
    :param doc: Optional documentation string for the property.
    :raises: FontPartsError: If the getter or setter method is not defined.

    Example of why this is needed:

    .. code-block:: python

        class BaseObject:

            _foo = 1

            def _get_foo(self):
                return self._foo

            def _set_foo(self, value):
                self._foo = value

            foo = property(_get_foo, _set_foo)


        class MyObject(BaseObject):

            def _set_foo(self, value):
                self._foo = value * 100


        >>> m = MyObject()
        >>> m.foo
        1
        >>> m.foo = 2
        >>> m.foo
        2

    The expected value is ``200``. The ``_set_foo`` method needs to be
    reregistered. Doing that also requires reregistering the ``_get_foo``
    method. It's possible to do this, but it's messy and will make subclassing
    less than ideal.

    Using dynamicProperty solves this:

    .. code-block:: python

        class BaseObject(object):

            _foo = 1

            foo = dynamicProperty("foo")

            def _get_foo(self):
                return self._foo

            def _set_foo(self, value):
                self._foo = value


        class MyObject(BaseObject):

            def _set_foo(self, value):
                self._foo = value * 100


        >>> m = MyObject()
        >>> m.foo
        1
        >>> m.foo = 2
        >>> m.foo
        200

    """

    def __init__(self, name: str, doc: str | None = None) -> None:
        self.name = name
        self.__doc__ = doc
        self.getterName = "_get_" + name
        self.setterName = "_set_" + name

    def __get__(self, obj: Any, cls: type[Any]) -> Any:
        getter = getattr(obj, self.getterName, None)
        if getter is not None:
            return getter()
        else:
            # obj is None when the property is accessed
            # via the class instead of an instance
            if obj is None:
                return self
            raise FontPartsError(f"no getter for {self.name!r}")

    def __set__(self, obj: Any, value: Any) -> None:
        setter = getattr(obj, self.setterName, None)
        if setter is not None:
            setter(value)
        else:
            raise FontPartsError(f"no setter for {self.name!r}")


def interpolate(
    minValue: InterpolatableType,
    maxValue: InterpolatableType,
    factor: InterpolationFactorLike,
) -> InterpolatableType:
    """Interpolate between two number-like objects.

    This method performs linear interpolation, calculating a value that is
    proportionally between `minValue` and `maxValue`, determined by the factor `factor`.

    :param minValue: The value corresponding to the 0.0 position in the interpolation
        as any object :ref:`emulating numeric types <numeric-types>`.
    :param maxValue: The value corresponding to the 1.0 position in the interpolation
        as any object :ref:`emulating numeric types <numeric-types>`.
    :param factor: The factor value determining the interpolation between
        `minValue` and `maxValue` as a single :class:`int` or :class:`float`.
        If `minValue` and `maxValue` supports it, `factor` may also be
        a :class:`tuple` of two :class:`int` or :class:`float` values representing
        the factors ``(x, y)``.
    :return: The interpolated value as any object :ref:`emulating numeric types
        <numeric-types>`.
    :raises TypeError: If `minValue` or `maxValue` does not support the provided
        `factor` type.

    """
    try:
        return minValue + (maxValue - minValue) * factor
    except TypeError as exc:
        raise TypeError(
            f"Factor must be an int or minValue float, not {type(factor).__name__}."
        ) from exc


# ------------
# Base Objects
# ------------


class BaseObject(Generic[BaseObjectType]):
    r"""Provide common base functionality to objects.

    This class is intended to serve as a foundation for other classes, supplying
    essential behaviors like initialization, string representation, comparison,
    and copying, while leaving more specific implementations to be provided by
    subclasses.

    Subclasses are expected to override or extend methods to suit their own
    behavior.

    :cvar copyClass: The class used for copying, defaults to the subclass being copied.
    :cvar copyAttributes: A tuple of attribute names to be copied when calling `copyData`.
    :param \*args: Any positional arguments.
    :param \**kwargs: Any keyword arguments.

    """

    # --------------
    # Initialization
    # --------------

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._init(*args, **kwargs)

    def _init(self, *args: Any, **kwargs: Any) -> None:
        r"""Initialize the native object.

        This is the environment implementation of :meth:`BaseObject.__init__`.

        :param \*args: Any positional arguments.
        :param \**kwargs: Any keyword arguments.

        .. note::

            Subclasses may override this method.

        """

    # ----
    # repr
    # ----

    def __repr__(self) -> str:
        contents = self._reprContents()
        if contents:
            contentString = " ".join(contents)
            contentString = " " + contentString
        else:
            contentString = ""
        return f"<{self.__class__.__name__}{contentString} at {hex(id(self))}>"

    def _reprContents(self) -> list[str]:
        """Provide a list of strings for inclusion in :meth:`BaseObject.__repr__.

        :return: A :class:`list` of :class:`str` items.

        .. note::

            Subclasses may override this method.
            If so, they should call :class:`super` and append their additions.

        """
        return []

    # --------
    # equality
    # --------

    def __eq__(self, other: Any) -> bool:
        """Check for equality with another object.

        :param other: The object to compare with.
        :return: :obj:`True` if the objects are equal, :obj:`False` otherwise.

        .. note::

            Subclasses may override this method.

        """
        if isinstance(other, self.__class__):
            return self.naked() is other.naked()
        return NotImplemented

    def __ne__(self, other: Any) -> bool:
        """Check for inequality with another object.

        :param other: The object to compare with.
        :return: :obj:`True` if the objects are unequal, :obj:`False` otherwise.

        .. note::

            Subclasses must not override this method.

        """
        equal = self.__eq__(other)
        if equal is NotImplemented:
            return NotImplemented
        return not equal

    # ----
    # Hash
    # ----

    def __hash__(self) -> int:
        """Return the hash value for the object.

        This allows subclasses to be used in hashable collections such as sets
        and dictionaries.

        :return: The hash value for the object as an :class:`int`.

        .. note::

            Subclasses may override this method.

        """
        return id(self.naked())

    # ----
    # Copy
    # ----

    copyClass: type[Any] | None = None
    copyAttributes: tuple[str, ...] = ()

    def copy(self: BaseObjectType) -> BaseObjectType:
        """Copy the current object into a new object of the same type.

        The returned object will not have a parent object.

        :return: A new :class:`BaseObject` subclass instance with the same attributes.

        """
        copyClass = self.copyClass
        if copyClass is None:
            copyClass = self.__class__
        copied = copyClass()
        copied.copyData(self)
        return copied

    def copyData(self, source: BaseObjectType) -> None:
        """Copy data from `source` into the current object.

        .. note::

            Subclasses may override this method.
            If so, they should call the :class:`super`.

        """
        for attr in self.copyAttributes:
            selfValue = getattr(self, attr)
            sourceValue = getattr(source, attr)
            if isinstance(selfValue, BaseObject):
                selfValue.copyData(sourceValue)
            else:
                setattr(self, attr, deepcopy(sourceValue))

    # ----------
    # Exceptions
    # ----------

    def raiseNotImplementedError(self) -> NoReturn:
        """Raise a :class:`NotImplementedError`.

        This exception needs to be raised frequently by the :mod:`fontParts.base`
        classes, so, it's here for convenience.

        :raises NotImplementedError: Whenever this method is called.

        """
        raise NotImplementedError(
            f"The {self.__class__.__name__} subclass does not implement this method."
        )

    # ---------------------
    # Environment Fallbacks
    # ---------------------

    def changed(self, *args: Any, **kwargs: Any) -> None:
        r"""Tell the environment that something has changed in the object.

        The behavior of this method will vary from environment to environment.

        :param \*args: Any positional arguments.
        :param \**kwargs: Any keyword arguments.

        Example::

            >>> obj.changed()

        """

    def naked(self) -> Any:
        """Return the environment's native object wrapped by the current object.

        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        Example::

            >>> loweLevelObj = obj.naked()

        """
        self.raiseNotImplementedError()


class BaseItems(Generic[KeyType, ValueType]):
    """Provide the given mapping with an items view object.

    This class provides a view of the key-value pairs in a mapping, similar to
    the behavior of Python's :meth:`dict.items` method. The view dynamically
    reflects any changes made to the parent mapping object, ensuring it remains
    up-to-date.

    The normalization method :meth:`_normalizeItems` must be defined by the
    parent `mapping` class.

    :param mapping: The parent :class:`BaseDict` instance whose items are
        represented by this view.

    """

    def __init__(self, mapping: BaseDict[KeyType, ValueType]) -> None:
        self._mapping = mapping

    def __contains__(self, item: tuple[KeyType, ValueType]) -> bool:
        """Check if a key-value pair exists in the mapping.

        :param item: The key-value pair to check for existence as a :class:`tuple`.
        :return: :obj:`True` if the key-value pair exists, :obj:`False` otherwise.

        """
        key, value = item
        normalizedItem = (
            self._mapping._normalizeKey(key),
            self._mapping._normalizeValue(value),
        )
        return normalizedItem in self._mapping._normalizeItems()

    def __iter__(self) -> Iterator[tuple[KeyType, ValueType]]:
        """Return an iterator over the key-value pairs in the mapping.

        This method yields each item one by one, removing it from the list of
        keys after it is yielded.

        :return: An iterator over the mapping's key-value pairs.

        """
        return iter(self._mapping._normalizeItems())

    def __len__(self) -> int:
        """Return the number of keys-value pairs in the mapping.

        :return: An :class:`int` representing the number of mapping items.

        """
        return len(self._mapping._normalizeItems())

    def __repr__(self) -> str:
        """Return a string representation of the items view.

        :return: A representation of the keys view as a :class:`str`.

        """
        return f"{self._mapping.__class__.__name__}_items({list(self)})"


class BaseKeys(Generic[KeyType]):
    """Provide the given mapping with a keys view object.

    This class provides a view of the keys in a mapping, similar to the behavior
    of Python's :meth:`dict.keys` method. The view dynamically reflects any
    changes made to the parent mapping object, ensuring it remains up-to-date.

    The normalization method :meth:`_normalizeItems` must be defined by the
    parent `mapping` class.

    :param mapping: The parent :class:`BaseDict` instance whose keys are
        represented by this view.

    """

    def __init__(self, mapping: BaseDict[KeyType, Any]) -> None:
        self._mapping = mapping

    def __contains__(self, key: KeyType) -> bool:
        """Check if a key exists in the mapping.

        :param key: The key to check for existence.
        :return: :obj:`True` if the key exists, :obj:`False` otherwise.

        """
        return any(k == key for k, _ in self._mapping._normalizeItems())

    def __iter__(self) -> Iterator[KeyType]:
        """Return an iterator over the keys in the mapping.

        This method yields each key one by one, removing it from the list of
        keys after it is yielded.

        :return: An iterator over the mapping's keys.

        """
        return (k for k, _ in self._mapping._normalizeItems())

    def __len__(self) -> int:
        """Return the number of keys in the mapping.

        :return: An :class:`int` representing the number of mapping keys.

        """
        return len(self._mapping._normalizeItems())

    def __repr__(self) -> str:
        """Return a string representation of the keys view.

        :return: A representation of the keys view as a :class:`str`.

        """
        return f"{self._mapping.__class__.__name__}_keys({list(self)})"

    def isdisjoint(self, other: Iterable[KeyType]) -> bool:
        """Check if the keys view has no common elements with another iterable.

        :param other: The iterable to compare against.
        :return: :obj:`True` if there are no common elements, :obj:`False` otherwise.

        """
        return set(self).isdisjoint(other)


class BaseValues(Generic[ValueType]):
    """Provide the given mapping with a values view object.

    This class provides a view of the values in a mapping, similar to the behavior
    of Python's :meth:`dict.values` method. The view dynamically reflects any
    changes made to the parent mapping object, ensuring it remains up-to-date.

    The normalization method :meth:`_normalizeItems` must be defined by the
    parent `mapping` class.

    :param mapping: The parent :class:`BaseDict` instance whose keys are
        represented by this view.

    """

    def __init__(self, mapping: BaseDict[Any, ValueType]) -> None:
        self._mapping = mapping

    def __contains__(self, value: ValueType) -> bool:
        """Check if a value exists in the mapping.

        :param value: The value to check for existence.
        :return: :obj:`True` if the value exists, :obj:`False` otherwise.

        """
        return any(v == value for _, v in self._mapping._normalizeItems())

    def __iter__(self) -> Iterator[ValueType]:
        """Return an iterator over the values in the mapping.

        This method yields each value one by one, removing it from the list of
        values after it is yielded.

        :return: An iterator over the mapping's values.

        """
        return (v for _, v in self._mapping._normalizeItems())

    def __len__(self) -> int:
        """Return the number of values in the mapping.

        :return: An :class:`int` representing the number of mapping values.

        """
        return len(self._mapping._normalizeItems())

    def __repr__(self) -> str:
        """Return a string representation of the values view.

        :return: A representation of the values view as a :class:`str`.

        """
        return f"{self._mapping.__class__.__name__}_values({list(self)})"


class BaseDict(BaseObject, Generic[KeyType, ValueType]):
    """Provide objects with basic dictionary-like functionality.

    :cvar keyNormalizer: An optional normalizer function for keys.
    :cvar valueNormalizer: An optional normalizer function for values.

    """

    keyNormalizer: Callable[[KeyType], KeyType] | None = None
    valueNormalizer: Callable[[ValueType], ValueType] | None = None

    def _normalizeKey(self, key: KeyType) -> KeyType:
        keyNormalizer = type(self).keyNormalizer
        return keyNormalizer(key) if keyNormalizer is not None else key

    def _normalizeValue(self, value: ValueType) -> ValueType:
        valueNormalizer = type(self).valueNormalizer
        return valueNormalizer(value) if valueNormalizer is not None else value

    def _normalizeItems(self) -> list[tuple[KeyType, ValueType]]:
        items = self._items()
        return [(self._normalizeKey(k), self._normalizeValue(v)) for (k, v) in items]

    def copyData(self, source: BaseObjectType) -> None:
        """Copy data from another object instance.

        This method calls the superclass's :meth:`copyData` method and updates
        the current :class:`BaseDict` instance with the contents of the `source`,
        provided it qualifies as a :class:`MutableMapping`.

        param source: The source :class:`BaseObject` instance from which
            to copy data.

        """
        super().copyData(source)
        if isinstance(source, MutableMapping):
            self.update(source)

    def __len__(self) -> int:
        """Return the number of items in the object.

        :return: An :class:`int` representing the number of dictionary items.

        """
        value = self._len()
        return value

    def _len(self) -> int:
        """Return the number of items in the native object.

        This is the environment implementation of :attr:`BaseDict.__len__`.

        :return: An :class:`int` representing the number of dictionary items.

        .. note::

            Subclasses may override this method.

        """
        return len(self.keys())

    def keys(self) -> BaseKeys[KeyType]:
        """Return a view of the keys in the object.

        :return: A :class:`BaseKeys` object instance of :class:`str` items.

        """
        return self._keys()

    def _keys(self) -> BaseKeys[KeyType]:
        """Return a view of the keys in the native object.

        This is the environment implementation of :meth:`BaseDict.keys`.

        :return: A :class:`BaseKeys` object instance of :class:`str` items. If
            a :cvar:`BaseDict.keyNormalizer` is set, it will be applied to each
            key in the returned view.

        .. note::

            Subclasses may override this method.

        """
        return BaseKeys(self)

    def items(self) -> BaseItems:
        """Return a view of the key-value pairs in the object.

        :return: A :class:`BaseItems` object instance.

        """
        return BaseItems(self)

    def _items(self) -> ItemsView[KeyType, ValueType]:
        """Return a view of the key-value pairs in the native object.

        This is the environment implementation of :meth:`BaseDict.items`.

        :return: A :class:`ItemsView` object instance.
            If both :cvar:`BaseDict.keyNormalizer` and :cvar:`BaseDict.valueNormalizer`
            are set, they will be applied to each key and value in the returned object.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def values(self) -> BaseValues[ValueType]:
        """Return a view of the values in the object.

        :return: A :class:`BaseValues` object instance.

        """
        return self._values()

    def _values(self) -> BaseValues[ValueType]:
        """Return a view of the values in the native object.

        This is the environment implementation of :meth:`BaseDict.values`.

        :return: A :class:`BaseValues` object instance. If
            a :cvar:`BaseDict.valueNormalizer` is set, it will be applied to each
            value in the returned view.

        .. note::

            Subclasses may override this method.

        """
        return BaseValues(self)

    def __contains__(self, key: KeyType) -> bool:
        """Check if a key is in the object.

        :param key: The key to check for.
        :return: :obj:`True` if the key is present, :obj:`False` otherwise.

        """
        key = self._normalizeKey(key)
        return self._contains(key)

    def _contains(self, key: KeyType) -> bool:
        """Check if a key is in the native object.

        This is the environment implementation of :meth:`BaseDict.__contains__`.

        :param key: The key to check for. If a :cvar:`BaseDict.keyNormalizer`
            is set, it will have been applied to the key before checking.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    has_key = __contains__

    def __setitem__(self, key: KeyType, value: ValueType) -> None:
        """Set the value for a given key in the object.

        :param key: The key to set.
        :param value: The value to set for the given key.

        """
        key = self._normalizeKey(key)
        value = self._normalizeValue(value)
        self._setItem(key, value)

    def _setItem(self, key: KeyType, value: ValueType) -> None:
        """Set the value for a given key in the native object.

        This is the environment implementation of :meth:`BaseDict.__setitem__`.

        :param key: The key to set. If a :cvar:`BaseDict.keyNormalizer`
            is set, it will have been applied to the value in the calling method.
        :param value: The value to set for the given key. If
            a :cvar:`BaseDict.valueNormalizer` is set, it will have been applied
            to the value in the calling method .
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def __getitem__(self, key: KeyType) -> ValueType:
        """Get the value for a given key from the object.

        :param key: The key to retrieve the value for.
        :return: The value for the given key.

        """
        key = self._normalizeKey(key)
        value = self._getItem(key)
        return self._normalizeValue(value)

    def _getItem(self, key: KeyType) -> ValueType:
        """Get the value for a given key from the native object.

        This is the environment implementation of :meth:`BaseDict.__getitem__`.

        :param key: The key to retrieve the value for. If
            a :cvar:`BaseDict.keyNormalizer` is set, it will have been applied to
            the key in the calling method.
        :return: The value for the given key. If
            a :cvar:`BaseDict.valueNormalizer` is set, it will be applied in the
            calling method to the returned value.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def get(self, key: KeyType, default: ValueType | None = None) -> ValueType | None:
        """Get the value for a given key in the object.

        If the given key is not found, The specified `default` will be returned.

        :param key: The key to look up.
        :param default: The default value to return if the key is not found.
            Defaults to :obj:`None`.
        :return: The value for the given key, or the default value if the key is
            not found.

        """
        key = self._normalizeKey(key)
        default = self._normalizeValue(default) if default is not None else default
        value = self._get(key, default=default)
        if value is not None and value is not default:
            value = self._normalizeValue(value)
        return value

    def _get(self, key: KeyType, default: ValueType | None) -> ValueType | None:
        """Get the value for a given key in the native object.

        This is the environment implementation of :meth:`BaseDict.get`.

        :param key: The key to look up. If a :cvar:`BaseDict.keyNormalizer` is
            set, it will have been applied to the given key.
        :param default: The default value to return if the key is not found.
        :return: The value associated with the given key, or the default value
            if the key is not found. If a :cvar:`BaseDict.valueNormalizer` is set,
            it will be applied in the calling method to the returned value.

        .. note::

            Subclasses may override this method.

        """
        if key in self:
            return self[key]
        return default

    def __delitem__(self, key: KeyType) -> None:
        """Delete a key-value pair from the object.

        :param key: The key to delete.

        """
        key = self._normalizeKey(key)
        self._delItem(key)

    def _delItem(self, key: KeyType) -> None:
        """Delete a key-value pair from the native object.

        This is the environment implementation of :meth:`BaseDict.__delitem__`.

        :param key: The key to delete. If a :cvar:`BaseDict.keyNormalizer` is
            set, it will have been applied to the given key.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def pop(self, key: KeyType, default: ValueType | None = None) -> ValueType | None:
        """Remove a key from the object and return it's value.

        If the given key is not found, The specified `default` will be returned.

        :param key: The key to remove.
        :param default: The optional default value to return if the key is not found.
            Defaults to :obj:`None`.
        :return: The value associated with the given key, or the default value
            if the key is not found.

        """
        key = self._normalizeKey(key)
        if default is not None:
            default = self._normalizeValue(default)
        value = self._pop(key, default=default)
        if value is not None:
            value = self._normalizeValue(value)
        return value

    def _pop(self, key: KeyType, default: ValueType | None) -> ValueType | None:
        """Remove a key from the native object and return it's value.

        This is the environment implementation of :meth:`BaseDict.pop`.

        :param key: The key to remove. If a :cvar:`BaseDict.keyNormalizer` is
            set, it will have been applied to the given key.
        :param default: The default value to return if the key is not found.
        :return: The value associated with the given key, or the default value
            if the key is not found. If a :cvar:`BaseDict.valueNormalizer` is set,
            it will be applied in the calling method to the returned value.

        .. note::

            Subclasses may override this method.

        """
        value = default
        if key in self:
            value = self[key]
            del self[key]
        return value

    def __iter__(self) -> Iterator[KeyType]:
        """Return an iterator over the keys of the object.

        This method yields each key one by one, removing it from the list of
        keys after it is yielded.

        :return: An iterator over the object's keys.

        """
        return self._iter()

    def _iter(self) -> Iterator[KeyType]:
        """Return an iterator over the keys of the native object.

        This is the environment implementation of :meth:`BaseDict.__iter__`.

        :return: An iterator over the object's keys.

        .. note::

            Subclasses may override this method.

        """
        yield from self.keys()

    def update(self, other: MutableMapping[KeyType, ValueType]) -> None:
        """Update the current object instance with key-value pairs from another.

        :param other: A :class:`MutableMapping` of key-value pairs to update
            this dictionary with.

        """
        otherCopy = deepcopy(other)
        d = {}
        for key, value in otherCopy.items():
            key = self._normalizeKey(key)
            value = self._normalizeValue(value)
            d[key] = value
        otherCopy = d
        self._update(otherCopy)

    def _update(self, other: MutableMapping[KeyType, ValueType]) -> None:
        """Update the current native object instance with key-value pairs from another.

        This is the environment implementation of :meth:`BaseDict.update`.

        :param other: A :class:`MutableMapping` of key-value pairs to update
            this dictionary with. If both :cvar:`BaseDict.keyNormalizer`
            and :cvar:`BaseDict.valueNormalizer` are set, they will have been
            applied to the keys and values, respectively.

        .. note::

            Subclasses may override this method.

        """
        for key, value in other.items():
            self[key] = value

    def clear(self) -> None:
        """Remove all items from the object."""
        self._clear()

    def _clear(self) -> None:
        """Remove all items from the native object.

        This is the environment implementation of :meth:`BaseDict.clear`.

        .. note::

            Subclasses may override this method.

        """
        for key in self.keys():
            del self[key]


class TransformationMixin(ABC):
    """Provide objects transformation-related functionality."""

    # ---------------
    # Transformations
    # ---------------

    def transformBy(
        self, matrix: AffineTransformationLike, origin: CoordinateLike | None = None
    ) -> None:
        """Transform the object according to the given matrix.

        :param matrix: The :ref:`type-transformation` to apply.
        :param origin: The optional point at which the transformation
            should originate as a:ref:`type-coordinate`. Defaults to :obj:`None`,
            representing an origin of ``(0, 0)``.

        Example::

            >>> obj.transformBy((0.5, 0, 0, 2.0, 10, 0))
            >>> obj.transformBy((0.5, 0, 0, 2.0, 10, 0), origin=(500, 500))

        """
        matrix = normalizers.normalizeTransformationMatrix(matrix)
        if origin is None:
            origin = (0, 0)
        origin = normalizers.normalizeCoordinateTuple(origin)
        if origin is not None:
            t = transform.Transform()
            oX, oY = origin
            t = t.translate(oX, oY)
            t = t.transform(matrix)
            t = t.translate(-oX, -oY)
            matrix = tuple(t)
        self._transformBy(matrix)

    def _transformBy(self, matrix: AffineTransformationLike, **kwargs: Any) -> None:
        r"""Transform the native object according to the given matrix.

        This is the environment implementation of :meth:`TransformationMixin.transformBy`.

        :param matrix: The :ref:`type-transformation` to apply. The value will
            have been normalized with
            :func:`normalizers.normalizeTransformationMatrix`.
        :param \**kwargs: Additional keyword arguments.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def moveBy(self, value: CoordinateLike) -> None:
        """Move the object according to the given coordinates.

        :param value: The x and y values to move the object by as
            a :ref:`type-coordinate`.

        Example::

            >>> obj.moveBy((10, 0))

        """
        value = normalizers.normalizeTransformationOffset(value)
        self._moveBy(value)

    def _moveBy(self, value: CoordinateLike, **kwargs: Any) -> None:
        r"""Move the native object according to the given coordinates.

        This is the environment implementation of :meth:`BaseObject.moveBy`.

        :param value: The x and y values to move the object by as
            a :ref:`type-coordinate`. The value will have been normalized with
            :func:`normalizers.normalizeTransformationOffset`.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        x, y = value
        t = transform.Offset(x, y)
        self.transformBy(tuple(t), **kwargs)

    def scaleBy(
        self, value: ScaleFactorLike, origin: CoordinateLike | None = None
    ) -> None:
        """Scale the object according to the given values.

        :param value: The value to scale the glyph by as a single :class:`int`
            or :class:`float`, or a :class:`tuple` or :class:`list` of
            two :class:`int` or :class:`float` values representing the values
            ``(x, y)``.
        :param origin: The optional point at which the scale should originate as
            a :ref:`type-coordinate`. Defaults to :obj:`None`, representing an
            origin of ``(0, 0)``.

        Example::

            >>> obj.scaleBy(2.0)
            >>> obj.scaleBy((0.5, 2.0), origin=(500, 500))

        """
        value = normalizers.normalizeTransformationScale(value)
        if origin is None:
            origin = (0, 0)
        origin = normalizers.normalizeCoordinateTuple(origin)
        self._scaleBy(value, origin=origin)

    def _scaleBy(
        self, value: ScaleFactorPair, origin: CoordinateLike, **kwargs: Any
    ) -> None:
        r"""Scale the native object according to the given values.

        This is the environment implementation of :meth:`BaseObject.scaleBy`.

        :param value: The value to scale the glyph by as a single :class:`int`
            or :class:`float`, or a :class:`tuple` or :class:`list` of
            two :class:`int` or :class:`float` values representing the values
            ``(x, y)``. The value will have been normalized
            with :func:`normalizeTransformationScale`.
        :param origin: The point at which the scale should originate as
            a :ref:`type-coordinate` or :obj:`None`. The value will have been
            normalized with :func:`normalizers.normalizeCoordinateTuple`.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        x, y = value
        t = transform.Identity.scale(x=x, y=y)
        self.transformBy(tuple(t), origin=origin, **kwargs)

    def rotateBy(
        self, value: IntFloatType, origin: CoordinateLike | None = None
    ) -> None:
        """Rotate the object by the specified value.

        :param value: The angle at which to rotate the object as an :class:`int`
            or a :class:`float`.
        :param origin: The optional point at which the rotation should originate
            as a :ref:`type-coordinate`. Defaults to :obj:`None`, representing an
            origin of ``(0, 0)``.

        Example::

            >>> obj.rotateBy(45)
            >>> obj.rotateBy(45, origin=(500, 500))

        """
        value = normalizers.normalizeRotationAngle(value)
        if origin is None:
            origin = (0, 0)
        origin = normalizers.normalizeCoordinateTuple(origin)
        self._rotateBy(value, origin=origin)

    def _rotateBy(self, value: float, origin: CoordinateLike, **kwargs: Any) -> None:
        r"""Rotate the native object by the specified value.

        This is the environment implementation of :meth:`TransformationMixin.rotateBy`.

        :param value: The angle at which to rotate the object as an :class:`int`
            or a :class:`float`. The value will have been normalized with
            :func:`normalizers.normalizeRotationAngle`.
        :param origin: The point at which the rotation should originate as
            a :ref:`type-coordinate` or :obj:`None`. The value will have been
            normalized with :func:`normalizers.normalizeCoordinateTuple`.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        a = math.radians(value)
        t = transform.Identity.rotate(a)
        self.transformBy(tuple(t), origin=origin, **kwargs)

    def skewBy(
        self, value: SkewAngleLike, origin: CoordinateLike | None = None
    ) -> None:
        """Skew the object by the given value.

        :param value: The value by which to skew the object as either a
            single :class:`int` or :class:`float` corresponding to the x
            direction, or a :class:`tuple` of two :class:`int` or :class:`float`
            values corresponding to the x and y directions.
        :param origin: The optional point at which the rotation should originate
            as a :ref:`type-coordinate`. Defaults to :obj:`None`, representing an
            origin of ``(0, 0)``.

        Example::

            >>> obj.skewBy(11)
            >>> obj.skewBy((25, 10), origin=(500, 500))

        """
        value = normalizers.normalizeTransformationSkewAngle(value)
        if origin is None:
            origin = (0, 0)
        origin = normalizers.normalizeCoordinateTuple(origin)
        self._skewBy(value, origin=origin)

    def _skewBy(
        self, value: SkewAnglePair, origin: CoordinateLike, **kwargs: Any
    ) -> None:
        r"""Skew the native object by the given value.

        This is the environment implementation of :meth:`TransformationMixin.skewBy`.

        :param value: The value by which to skew the object as either a
            single :class:`int` or :class:`float` corresponding to the x
            direction, or a :class:`tuple` of two :class:`int` or :class:`float`
            values corresponding to the x and y directions. The value will have
            been normalized with :func:`normalizers.normalizeTransformationSkewAngle`.
        :param origin: The point at which the rotation should originate
            as a :ref:`type-coordinate` or :obj:`None`. The value will have been
            normalized with :func:`normalizers.normalizeCoordinateTuple`.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        x, y = value
        x = math.radians(x)
        y = math.radians(y)
        t = transform.Identity.skew(x=x, y=y)
        self.transformBy(tuple(t), origin=origin, **kwargs)

    # ----------------
    # Abstract members
    # ----------------

    @abstractmethod
    def raiseNotImplementedError(self):
        pass


class InterpolationMixin(ABC):
    """Provide objects with interpolation-related functionality.

    :cvar compatibilityReporterClass:  A class used for reporting interpolation
        compatibility between two objects. If :obj:`None`, compatibility
        reporting is not enabled.

    """

    # -------------
    # Compatibility
    # -------------

    compatibilityReporterClass: type[Any] | None = None

    def isCompatible(self, other: Any, cls: type[Any]) -> tuple[bool, Any]:
        """Evaluate interpolation compatibility with another object.

        :param other: The other object instance to check compatibility with.
        :param cls: The class type to check the `other` object against.
        :return: A :class:`tuple` where the first element is a :class:`bool`
            indicating whether the objects are compatible, and the second
            element is the compatibility reporter instance.
        :raises TypeError: If `other` is not an instance of `cls`.

        """
        if not isinstance(other, cls):
            raise TypeError(
                f"""Compatibility between an instance of {cls.__name__!r}
                and an instance of {other.__class__.__name__!r} can not be checked."""
            )
        if self.compatibilityReporterClass is None:
            raise ValueError("The compatibility reporter class cannot be None.")
        reporter = self.compatibilityReporterClass(self, other)
        self._isCompatible(other, reporter)
        return not reporter.fatal, reporter

    def _isCompatible(self, other: Any, reporter: Any) -> None:
        """Evaluate interpolation compatibility with another native object.

        This is the environment implementation of :meth:`InterpolationMixin.isCompatible`.

        :param other: The other object instance to check compatibility with.
        :param reporter: An object used to report compatibility issues.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # ----------------
    # Abstract members
    # ----------------

    @abstractmethod
    def raiseNotImplementedError(self):
        pass


class SelectionMixin(ABC):
    """Provide objects with selection-related functionality."""

    # -------------
    # Selected Flag
    # -------------

    selected: dynamicProperty = dynamicProperty(
        "base_selected",
        """Get or set the object's selection state.

        The value must be a :class:`bool` indicating whether the object is
        selected or not.

        :return: :obj:`True` if the object is selected, :obj:`False` otherwise.

        Example::

            >>> obj.selected
            False
            >>> obj.selected = True

        """,
    )

    def _get_base_selected(self) -> bool:
        value = self._get_selected()
        value = normalizers.normalizeBoolean(value)
        return value

    def _set_base_selected(self, value: bool) -> None:
        value = normalizers.normalizeBoolean(value)
        self._set_selected(value)

    def _get_selected(self) -> bool:  # type: ignore[return]
        """Get or the object's selection state.

        This is the environment implementation
        of :attr:`BaseObject.selected` property getter.

        :return: :obj:`True` if the object is selected, :obj:`False` otherwise.
            The value will be normalized with :func:`normalizers.normalizeBoolean`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method if they implement object selection.

        """
        self.raiseNotImplementedError()

    def _set_selected(self, value: bool) -> None:
        """Set the object's selection state.

        This is the environment implementation of the
        :attr:`SelectionMixin.selected` property setter.

        :param value: Whether the object is selected. The value will have been
            normalized with :func:`normalizers.normalizeBoolean`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method if they implement object selection.

        """
        self.raiseNotImplementedError()

    # -----------
    # Sub-Objects
    # -----------
    @classmethod
    def _getSelectedSubObjects(cls, subObjects: Any) -> tuple[Any]:
        selected = tuple(obj for obj in subObjects if obj.selected)
        return selected

    @classmethod
    def _setSelectedSubObjects(cls, subObjects: Any, selected: Any) -> None:
        for obj in subObjects:
            obj.selected = obj in selected

    # ----------------
    # Abstract members
    # ----------------

    @abstractmethod
    def raiseNotImplementedError(self):
        pass


class PointPositionMixin(ABC):
    """Provide objects with the ability to determine point position.

    This class adds a `position` attribute as a :class:`dyanmicProperty`, for
    use as a mixin with objects that have `x` and `y` attributes.

    """

    position: dynamicProperty = dynamicProperty(
        "base_position",
        """Get or set the point position of the object.

        The value must be a :ref:`type-coordinate`.

        :return: The current point position as a :ref:`type-coordinate`.

        """,
    )

    def _get_base_position(self) -> Coordinate:
        value = self._get_position()
        value = normalizers.normalizeCoordinateTuple(value)
        return value

    def _set_base_position(self, value: CoordinateLike) -> None:
        value = normalizers.normalizeCoordinateTuple(value)
        self._set_position(value)

    def _get_position(self) -> Coordinate:
        """Get the point position of the object.

        This is the environment implementation of
        the :attr:`PointPositionMixin.position` property getter.

        :return: The current point position as a :ref:`type-coordinate`.
            The value will be normalized
            with :func:`normalizers.normalizeCoordinateTuple`.

        .. note::

            Subclasses may override this method.

        """
        return (self.x, self.y)

    def _set_position(self, value: CoordinateLike) -> None:
        """Set the point position of the object.

        This is the environment implementation of
        the :attr:`PointPositionMixin.position` property setter.

        :param value: The point position to set as a :ref:`type-coordinate`.
            The value will have been normalized
            with :func:`normalizers.normalizeCoordinateTuple`.

        .. note::

            Subclasses may override this method.

        """
        pX, pY = self.position
        x, y = value
        dX = x - pX
        dY = y - pY
        self.moveBy((dX, dY))

    # ----------------
    # Abstract members
    # ----------------

    x: dynamicProperty = dynamicProperty("base_x")

    @abstractmethod
    def _get_base_x(self) -> IntFloatType:
        pass

    @abstractmethod
    def _set_base_x(self, value: IntFloatType) -> None:
        pass

    @abstractmethod
    def _get_x(self) -> IntFloatType:
        pass

    @abstractmethod
    def _set_x(self, value: IntFloatType) -> None:
        pass

    y: dynamicProperty = dynamicProperty("base_y")

    @abstractmethod
    def _get_base_y(self) -> IntFloatType:
        pass

    @abstractmethod
    def _set_base_y(self, value: IntFloatType) -> None:
        pass

    @abstractmethod
    def _get_y(self) -> IntFloatType:
        pass

    @abstractmethod
    def _set_y(self, value: IntFloatType) -> None:
        pass

    @abstractmethod
    def moveBy(self, value):
        pass

    @abstractmethod
    def raiseNotImplementedError(self):
        pass


class IdentifierMixin(ABC):
    """Provide objects with a unique identifier."""

    # identifier

    identifier: dynamicProperty = dynamicProperty(
        "base_identifier",
        """Get the object's unique identifier.

        This attribute is read-only. Use :meth:`IdentifierMixin.getIdentifier`
        to request an identifier if it does not exist.

        :return: The unique identifier assigned to the object as a :class:`str`,
            or :obj:`None` indicating the object has no identifier.

        Example ::

            >>> object.identifier
            'ILHGJlygfds'

        """,
    )

    def _get_base_identifier(self) -> str | None:
        value = self._get_identifier()
        if value is not None:
            value = normalizers.normalizeIdentifier(value)
        return value

    def _get_identifier(self) -> str | None:  # type: ignore[return]
        """Get the native object's unique identifier.

        This is the environment implementation of :attr:`IdentifierMixin.identifier`.

        If the native object does not have an identifier assigned, one may be
        assigned with :meth:`IdentifierMixin.getIdentifier`

        :return: The unique identifier assigned to the object as a :class:`str`,
            or :obj:`None` indicating the object has no identifier.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def getIdentifier(self) -> str:
        """Generate and assign a unique identifier to the object.

        If the object already has an identifier, the existing one should
        be returned.

        :return: A unique object identifier as a :class:`str`.

        """
        return self._getIdentifier()

    def _getIdentifier(self) -> str:  # type: ignore[return]
        """Generate and assign a unique identifier to the native object.

        This is the environment implementation of :meth:`IdentifierMixin.getIdentifier`.

        :return: A unique object identifier as a :class:`str`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def _setIdentifier(self, value: str) -> None:
        """Force a specific identifier onto an object.

        This method is used internally to force a specific identifier onto an
        object in certain situations.

        :param value: The identifier to set as a :class:`str` or :obj:`None`.

        .. note::

            Subclasses that allow setting an identifer may override this method.

        """
        pass

    # ----------------
    # Abstract members
    # ----------------

    @abstractmethod
    def raiseNotImplementedError(self):
        pass


def reference(obj: Any) -> Callable[[], Any]:
    """
    This code returns a simple function that returns the given object.
    This is a backwards compatibility function that is under review (see issue #749).
    We used to use weak references, but they proved problematic (see issue #71),
    so this function was put in place to make sure existing code continued to
    function. The need for it is questionable, so it may be deleted soon.
    """

    def wrapper() -> Any:
        return obj

    return wrapper


class FuzzyNumber:
    """Represent a number like object with a threshold.

    This class should be used to compare numbers where a threshold is needed.

    """

    def __init__(self, value: IntFloatType, threshold: IntFloatType) -> None:
        self.value = value
        self.threshold = threshold

    def __repr__(self) -> str:
        return f"[{self.value:f} {self.threshold:f}]"

    def __lt__(self, other: FuzzyNumber | IntFloatType) -> bool:
        if hasattr(other, "value"):
            if abs(self.value - other.value) < self.threshold:
                return False
            else:
                return self.value < other.value
        return self.value < other

    def __eq__(self, other: object) -> bool:
        if hasattr(other, "value"):
            return abs(self.value - other.value) < self.threshold
        return self.value == other

    def __hash__(self) -> int:
        return hash((self.value, self.threshold))
